import asyncio
import contextlib
import functools
import subprocess
import threading
import time
from concurrent import futures
from typing import Any, Awaitable, Callable, List, Mapping, Optional

import vcsinfo
from absl import logging

from lxm3._vendor.xmanager import xm
from lxm3._vendor.xmanager.xm import async_packager
from lxm3._vendor.xmanager.xm import core
from lxm3._vendor.xmanager.xm import id_predictor
from lxm3._vendor.xmanager.xm import job_blocks
from lxm3._vendor.xmanager.xm import pattern_matching as pm
from lxm3.xm_cluster import config as config_lib
from lxm3.xm_cluster import executors
from lxm3.xm_cluster import metadata
from lxm3.xm_cluster import packaging
from lxm3.xm_cluster.console import console
from lxm3.xm_cluster.execution import gridengine as gridengine_execution
from lxm3.xm_cluster.execution import local as local_execution
from lxm3.xm_cluster.execution import slurm as slurm_execution


def _gridengine_job_predicate(job: xm.Job):
    return isinstance(job.executor, executors.GridEngine)


def _slurm_job_predicate(job: xm.Job):
    return isinstance(job.executor, executors.Slurm)


def _local_job_predicate(job: xm.Job):
    return isinstance(job.executor, executors.Local)


class _LaunchResult:
    def __init__(self, local_handles, non_local_handles):
        self.local_handles = local_handles
        self.non_local_handles = non_local_handles


async def _launch(jobs: List[xm.Job]):
    experiment: ClusterExperiment = core._current_experiment.get()  # type: ignore
    gridengine_jobs = list(filter(_gridengine_job_predicate, jobs))
    slurm_jobs = list(filter(_slurm_job_predicate, jobs))
    local_jobs = list(filter(_local_job_predicate, jobs))

    if not (
        len(gridengine_jobs) == len(jobs)
        or len(local_jobs) == len(jobs)
        or len(slurm_jobs) == len(jobs)
    ):
        raise ValueError(
            "ClusterExperiment only supports running only GridEngine XOR "
            "Local XOR Slurm executors at the same time."
        )

    local_handles = []
    if local_jobs:
        local_handles.extend(
            await local_execution.launch(experiment._config, local_jobs)  # type: ignore
        )

    non_local_handles = []
    if gridengine_jobs:
        non_local_handles.extend(
            await gridengine_execution.launch(experiment._config, gridengine_jobs)  # type: ignore
        )

    if slurm_jobs:
        non_local_handles.extend(
            await slurm_execution.launch(experiment._config, slurm_jobs)  # type: ignore
        )

    return _LaunchResult(local_handles, non_local_handles)


class ClusterWorkUnit(xm.WorkUnit):
    """A mock version of WorkUnit with abstract methods implemented."""

    experiment: "ClusterExperiment"

    def __init__(
        self,
        experiment: "ClusterExperiment",
        work_unit_id_predictor: id_predictor.Predictor,
        create_task: Callable[[Awaitable[Any]], futures.Future[Any]],
        launched_jobs: List[job_blocks.JobType],
        launched_jobs_args: List[Optional[Mapping[str, Any]]],
        args: Optional[Mapping[str, Any]],
        role: xm.ExperimentUnitRole,
    ) -> None:
        super().__init__(experiment, create_task, args, role)
        self._launched_jobs = launched_jobs
        self._launched_jobs_args = launched_jobs_args
        self._work_unit_id = work_unit_id_predictor.reserve_id()
        self._work_unit_id_predictor = work_unit_id_predictor
        self._launch_event = threading.Event()
        self._local_handles = []
        self._non_local_handles = []

    async def _launch_job_group(
        self,
        job_group: xm.JobGroup,
        args: Optional[Mapping[str, Any]],
        identity: str,
    ) -> None:
        """Appends the job group to the launched_jobs list."""
        del identity

        async with self._work_unit_id_predictor.submit_id(self._work_unit_id):  # type: ignore
            await self._submit_job_for_execution(job_group, args)

        # This is used by batched experiment to wait for all jobs to be launched.
        # before initiating a batch context
        self._launch_event.set()

    async def _submit_job_for_execution(self, job_group: xm.JobGroup, args):
        jobs: List[xm.Job] = list(job_group.jobs.values())  # type: ignore
        assert len(jobs) == 1
        if self.experiment.is_in_batch():

            def callback(result):
                self._ingest_handles(result)

            assert len(jobs) == 1
            self.experiment._register_delayed_job((jobs[0], args), callback)
        else:
            launch_result = await _launch(jobs)
            self._ingest_handles(launch_result)

    def _ingest_handles(self, launch_result):
        """"""
        self._local_handles.extend(launch_result.local_handles)
        self._non_local_handles.extend(launch_result.non_local_handles)

    async def wait_for_local_jobs(self, is_exit_abrupt: bool):
        if not is_exit_abrupt:
            await asyncio.gather(*[handle.wait() for handle in self._local_handles])

    @property
    def work_unit_id(self) -> int:
        return self._work_unit_id

    @property
    def experiment_unit_name(self) -> str:
        return f"{self.experiment_id}_{self._work_unit_id}"

    @property
    def context(self) -> metadata.ClusterMetadataContext:
        return metadata.ClusterMetadataContext()


class ClusterExperiment(xm.Experiment):
    """A mock version of Experiment with abstract methods implemented."""

    _async_packager = async_packager.AsyncPackager(packaging.package)

    def __init__(
        self,
        experiment_title: str,
        config: Mapping[str, Any],
        vcs: Optional[vcsinfo.VCS] = None,
    ) -> None:
        super().__init__()
        self.launched_jobs = []
        self.launched_jobs_args = []
        self._work_units = []
        self._experiment_id = int(time.time() * 10**3)
        self._in_batch_lock = threading.Lock()
        self._in_batch = False
        self.delayed_jobs = []
        self._experiment_title = experiment_title
        self._config = config
        self._vcs = vcs

    def is_in_batch(self):
        with self._in_batch_lock:
            return self._in_batch

    @contextlib.contextmanager
    def batch(self):
        is_coro_context = False
        try:
            asyncio.get_running_loop()
            is_coro_context = True
        except RuntimeError:
            pass
        if is_coro_context:
            raise RuntimeError(
                "Launching batch experiment from async context is not yet supported."
            )
        try:
            assert not self._in_batch
            if len(self._work_units) > 0:
                self._work_units[-1]._launch_event.wait()
            self._in_batch = True
            yield
        finally:
            if len(self._work_units) > 0:
                self._work_units[-1]._launch_event.wait()
            delayed_jobs = [j[0] for j in self.delayed_jobs]
            delayed_cb = [j[2] for j in self.delayed_jobs]

            async def launch_array_jobs():
                array_launch_result = await _launch(delayed_jobs)
                results = []
                if array_launch_result.local_handles:
                    results = [_LaunchResult(array_launch_result.local_handles, [])]
                else:
                    results = [_LaunchResult([], array_launch_result.non_local_handles)]

                for callback, results in zip(delayed_cb, results):
                    callback(results)

            self._create_task(launch_array_jobs())
            self.delayed_jobs = []
            self._in_batch = False

    def _register_delayed_job(self, job_and_args, callback):
        job, args = job_and_args
        self.delayed_jobs.append((job, args, callback))

    def _create_experiment_unit(
        self,
        args: Optional[Mapping[str, Any]],
        role: xm.ExperimentUnitRole = xm.WorkUnitRole(),
        identity: str = "",
    ) -> Awaitable[ClusterWorkUnit]:
        """Creates a new WorkUnit instance for the experiment."""
        del identity  # Unused.
        future = asyncio.Future(loop=self._event_loop)
        experiment_unit = ClusterWorkUnit(
            self,
            self._work_unit_id_predictor,
            self._create_task,
            self.launched_jobs,
            self.launched_jobs_args,
            args,
            role,
        )

        def _unsupported_aux_units(_):
            raise NotImplementedError("Auxiliary units are not supported")

        pm.match(
            pm.Case(
                [xm.WorkUnitRole],
                lambda _: self._work_units.append(experiment_unit),
            ),
            pm.Case(
                [xm.AuxiliaryUnitRole],
                _unsupported_aux_units,
            ),
        )(role)

        future.set_result(experiment_unit)
        return future

    def _wait_for_local_jobs(self, is_exit_abrupt: bool):
        if self._work_units:
            if any([wu._local_handles for wu in self._work_units]):
                console.print(
                    "Waiting for local jobs to complete. "
                    "Press Ctrl+C to terminate them and exit"
                )
        for unit in self._work_units:
            self._create_task(unit.wait_for_local_jobs(is_exit_abrupt))

    def __exit__(self, exc_type, exc_value, traceback):
        # Flush `.add` calls.
        self._wait_for_tasks()
        self._wait_for_local_jobs(exc_value is not None)
        return super().__exit__(exc_type, exc_value, traceback)

    async def __aexit__(self, exc_type, exc_value, traceback):
        # Flush `.add` calls.
        await self._await_for_tasks()
        self._wait_for_local_jobs(exc_value is not None)
        return await super().__aexit__(exc_type, exc_value, traceback)

    @property
    def work_unit_count(self) -> int:
        return len(self.work_units)

    @property
    def work_units(self):
        return self._work_units

    @property
    def experiment_id(self) -> int:
        return self._experiment_id

    @property
    def context(self) -> metadata.ClusterMetadataContext:
        return metadata.ClusterMetadataContext()


@functools.lru_cache()
def _load_vcsinfo() -> Optional[vcsinfo.VCS]:
    vcs = None

    try:
        vcs_root = subprocess.check_output(
            ["git", "rev-parse", "--show-toplevel"], text=True
        ).strip()
        vcs = vcsinfo.detect_vcs(vcs_root)
    except subprocess.SubprocessError:
        logging.debug("Failed to detect VCS info")

    return vcs


def create_experiment(
    experiment_title: str, config: Optional[config_lib.Config] = None
) -> ClusterExperiment:
    """Create a LXM3 experiment backed by the xm_cluster backend.
    Args:
        experiment_title: Title of the experiment.
        config: Optional config object to use. If set, override
            the configuration loaded from the config file.
    """
    config = config or config_lib.default()
    vcs = _load_vcsinfo()

    if not config.project() and vcs is not None:
        config.data["project"] = vcs.name

    return ClusterExperiment(experiment_title, config=config)



#!/usr/bin/env python3
from absl import app
from absl import flags

from lxm3 import xm
from lxm3 import xm_cluster

_SINGULARITY_CONTAINER = flags.DEFINE_string(
    "container", "jax-cuda.sif", "Path to singularity container"
)
_BATCH = flags.DEFINE_bool("batch", False, "If set, launch array job example")


def main(_):
    with xm_cluster.create_experiment(experiment_title="basic") as experiment:
        executor: xm_cluster.Slurm = xm_cluster.Slurm(walltime=10 * xm.Min)

        spec = xm_cluster.PythonPackage(
            # This is a relative path to the launcher that contains
            # your python package (i.e. the directory that contains pyproject.toml)
            path=".",
            # Entrypoint is the python module that you would like to
            # In the implementation, this is translated to
            #   python3 -m py_package.main
            entrypoint=xm_cluster.ModuleName("py_package.main"),
        )

        # Wrap the python_package to be executing in a singularity container.
        singularity_container = _SINGULARITY_CONTAINER.value

        # It's actually not necessary to use a container, without it, we
        # fallback to the current python environment for local executor and
        # whatever Python environment picked up by the cluster for GridEngine.
        # For remote execution, using the host environment is not recommended.
        # as you may spend quite some time figuring out dependency problems than
        # writing a simple Dockfiler/Singularity file.
        if singularity_container is not None:
            spec = xm_cluster.SingularityContainer(
                spec,
                image_path=singularity_container,
            )

        [executable] = experiment.package(
            [xm.Packageable(spec, executor_spec=executor.Spec())]
        )

        if not _BATCH.value:
            experiment.add(
                xm.Job(
                    executable=executable,
                    executor=executor,
                    # You can pass additional arguments to your executable with args
                    # This will be translated to `--seed 1`
                    # Note for booleans we currently use the absl.flags convention
                    # so {'gpu': False} will be translated to `--nogpu`
                    args={"seed": 1},
                    # You can customize environment_variables as well.
                    env_vars={"XLA_PYTHON_CLIENT_PREALLOCATE": "false"},
                )
            )
        else:
            with experiment.batch():
                for _ in range(2):
                    experiment.add(
                        xm.Job(
                            executable=executable,
                            executor=executor,
                            # You can pass additional arguments to your executable with args
                            # This will be translated to `--seed 1`
                            # Note for booleans we currently use the absl.flags convention
                            # so {'gpu': False} will be translated to `--nogpu`
                            args={"seed": 1},
                            # You can customize environment_variables as well.
                            env_vars={"XLA_PYTHON_CLIENT_PREALLOCATE": "false"},
                        )
                    )


if __name__ == "__main__":
    app.run(main)



import asyncio
import atexit
import concurrent.futures
import datetime
import os
import shutil
import subprocess
from typing import List, Optional

from absl import logging

from lxm3 import xm
from lxm3.xm_cluster import config as config_lib
from lxm3.xm_cluster import executables
from lxm3.xm_cluster import executors
from lxm3.xm_cluster.console import console
from lxm3.xm_cluster.execution import artifacts
from lxm3.xm_cluster.execution import common
from lxm3.xm_cluster.execution import job_script

_LOCAL_EXECUTOR: Optional[concurrent.futures.ThreadPoolExecutor] = None

_TASK_OFFSET = 1
_JOB_SCRIPT_SHEBANG = "#!/usr/bin/env bash"
_TASK_ID_VAR_NAME = "SGE_TASK_ID"


def local_executor():
    global _LOCAL_EXECUTOR
    if _LOCAL_EXECUTOR is None:
        _LOCAL_EXECUTOR = concurrent.futures.ThreadPoolExecutor(1)

        def shutdown():
            if _LOCAL_EXECUTOR is not None:
                logging.debug("Shutting down local executor...")
                _LOCAL_EXECUTOR.shutdown()

        atexit.register(shutdown)
    return _LOCAL_EXECUTOR


def _is_gpu_requested(executor: executors.Local) -> bool:
    return shutil.which("nvidia-smi") is not None


def _create_job_header(
    executor: executors.Local, jobs: List[xm.Job], job_script_dir: str, job_name: str
) -> str:
    return ""


def _get_setup_cmds(
    executable: executables.Command,
    executor: executors.Local,
) -> str:
    del executor
    cmds = ["echo >&2 INFO[$(basename $0)]: Running on host $(hostname)"]

    if executable.singularity_image is not None:
        cmds.append(
            "echo >&2 INFO[$(basename $0)]: Singularity version: $(singularity --version)"
        )
    return "\n".join(cmds)


def deploy_job_resources(
    artifact: artifacts.Artifact,
    jobs: List[xm.Job],
    version: Optional[str] = None,
) -> str:
    version = version or datetime.datetime.now().strftime("%Y%m%d.%H%M%S")

    executable = jobs[0].executable
    executor = jobs[0].executor

    assert isinstance(executor, executors.Local)
    assert isinstance(executable, executables.Command)
    job_script.validate_same_job_configuration(jobs)

    job_name = f"job-{version}"

    job_script_dir = artifact.job_path(job_name)

    setup = _get_setup_cmds(executable, executor)
    header = _create_job_header(executor, jobs, job_script_dir, job_name)

    return common.create_array_job(
        artifact=artifact,
        executable=executable,
        singularity_image=executable.singularity_image,
        singularity_options=executor.singularity_options,
        jobs=jobs,
        use_gpu=_is_gpu_requested(executor),
        version=version,
        job_script_shebang=_JOB_SCRIPT_SHEBANG,
        task_offset=_TASK_OFFSET,
        task_id_var_name=_TASK_ID_VAR_NAME,
        setup=setup,
        header=header,
    )


class LocalExecutionHandle:
    def __init__(self, future: concurrent.futures.Future) -> None:
        self.future = future

    async def wait(self) -> None:
        return await asyncio.wrap_future(self.future)

    async def monitor(self) -> None:
        await asyncio.wrap_future(self.future)


async def launch(config: config_lib.Config, jobs: List[xm.Job]):
    if len(jobs) < 1:
        return []

    local_config = config.local_config()
    artifact = artifacts.LocalArtifact(
        local_config["storage"]["staging"], project=config.project()
    )
    job_script_path = deploy_job_resources(artifact, jobs)

    console.print(f"Launching {len(jobs)} jobs locally...")
    handles = []
    for i in range(len(jobs)):

        def task(i):
            subprocess.run(
                ["bash", job_script_path],
                env={**os.environ, _TASK_ID_VAR_NAME: str(i + _TASK_OFFSET)},
            )

        future = local_executor().submit(task, i)
        handles.append(LocalExecutionHandle(future))

    return handles



# type: ignore

from absl import app

from lxm3 import xm
from lxm3 import xm_cluster


def main(_):
    with xm_cluster.create_experiment(experiment_title="universal") as experiment:
        executor = xm_cluster.Local()
        executable = xm.Packageable(
            xm_cluster.UniversalPackage(
                path=".",
                entrypoint=["python3", "main.py"],
                build_script="build.sh",
            ),
            xm_cluster.Local.Spec(),  # type: ignore
        )
        [executable] = experiment.package([executable])
        experiment.add(xm.Job(executable=executable, executor=executor))


if __name__ == "__main__":
    app.run(main)



#!/usr/bin/env python3
from absl import app

from lxm3 import xm
from lxm3 import xm_cluster


def main(_):
    with xm_cluster.create_experiment(experiment_title="sweep") as experiment:
        requirements = xm_cluster.JobRequirements()
        executor = xm_cluster.Local(requirements=requirements)

        packageable = xm.Packageable(
            executable_spec=xm_cluster.PythonPackage(
                # This is a relative path to the launcher that contains
                # your python package (i.e. the directory that contains pyproject.toml)
                path=".",
                # Entrypoint is the python module that you would like to
                # In the implementation, this is translated to
                #   python3 -m py_package.main
                entrypoint=xm_cluster.ModuleName("py_package.main"),
            ),
            executor_spec=executor.Spec(),
        )

        [executable] = experiment.package([packageable])

        # To submit parameter sweep by array jobs, you can use the batch context
        # Without the batch context, jobs with be submitted individually.
        with experiment.batch():
            for i in range(2):
                experiment.add(
                    xm.Job(
                        executable=executable,
                        executor=executor,
                        # You can pass additional arguments to your executable with args
                        # This will be translated to `--seed 1`
                        # Note for booleans we currently use the absl.flags convention
                        # so {'gpu': False} will be translated to `--nogpu`
                        args={"seed": i},
                        # You can customize environment_variables as well.
                        env_vars={"TASK": str(i)},
                    )
                )

        # You can also use a job generator.
        # This is useful for example in a few cases
        # 1. if you want to configure a working directory depending on the work unit id.
        # 2. You can to dynamically compute some additional args/env_vars.
        #    one use case is to compute environment variables to be passed to wandb.
        async def make_job(work_unit: xm.WorkUnit, **args) -> None:
            work_unit.add(
                xm.Job(
                    executable=executable,
                    executor=executor,
                    args={"seed": args["i"]},
                    env_vars={"TASK": work_unit.work_unit_id},
                )
            )

        with experiment.batch():
            for i in range(2):
                experiment.add(make_job, {"i": i})


if __name__ == "__main__":
    app.run(main)



import subprocess


class PackagingError(Exception):
    pass


def test():
    try:
        subprocess.run(["pip", "balkfds"], check=True, capture_output=True)
    except subprocess.CalledProcessError as e:
        raise PackagingError("Unable to package") from e


if __name__ == "__main__":
    test()



import shlex
from typing import Any, Dict, List, Optional, Sequence, Union

from lxm3 import xm
from lxm3._vendor.xmanager.xm import job_blocks
from lxm3.xm_cluster import executables
from lxm3.xm_cluster import executors

_DEFAULT_ARRAY_WRAPPER_SHEBANG = "#!/usr/bin/env bash"

ARRAY_WRAPPER_NAME = "array_wrapper.sh"
JOB_SCRIPT_NAME = "job.sh"

_ARRAY_WRAPPER_TEMPLATE = """\
%(shebang)s

TASK_OFFSET="%(task_offset)d"

E_NOARGS=85
if [ $# -ne "1" ]
then
  echo "Usage: `basename $0` index"
  exit $E_NOARGS
else
  TASK_ID=$(($1 - $TASK_OFFSET))
fi

# Set up environment variable for task
__array_set_env_vars() {
%(env_vars)s
}

# Print command arguments for task
__array_get_args() {
%(args)s
}

__array_set_env_vars $TASK_ID
%(cmd)s $(__array_get_args $TASK_ID)
exit $?
"""

_WorkItem = Dict[str, Any]


def _create_env_vars(env_vars_list: List[Dict[str, str]]) -> str:
    """Create the env_vars list."""
    lines = []
    first_keys = set(env_vars_list[0].keys())
    if not first_keys:
        return ":;"
    for env_vars in env_vars_list:
        if first_keys != set(env_vars.keys()):
            raise ValueError("Expect all environment variables to have the same keys")
    for key in first_keys:
        for task_id, env_vars in enumerate(env_vars_list):
            lines.append(
                '{key}_{task_id}="{value}"'.format(
                    key=key, task_id=task_id, value=env_vars[key]
                )
            )
        lines.append('{key}=$(eval echo \\$"{key}_$1")'.format(key=key))
        lines.append("export {key}".format(key=key))
    content = "\n".join(lines)
    return content


def _create_args(args_list: List[List[str]]) -> str:
    """Create the args list."""
    if not args_list:
        return ":;"
    lines = []
    for task_id, args in enumerate(args_list):
        args_str = " ".join([a for a in args])
        lines.append(
            'TASK_CMD_ARGS_{task_id}="{args_str}"'.format(
                task_id=task_id, args_str=args_str
            )
        )
    lines.append('TASK_CMD_ARGS=$(eval echo \\$"TASK_CMD_ARGS_$1")')
    lines.append("echo $TASK_CMD_ARGS")
    content = "\n".join(lines)
    return content


def worklist_from_jobs(
    executable: executables.Command, jobs: List[xm.Job]
) -> List[_WorkItem]:
    work_list = []
    for job in jobs:
        work_list.append(
            {
                "args": job_blocks.merge_args(executable.args, job.args).to_list(
                    xm.utils.ARG_ESCAPER
                ),
                "env_vars": {**executable.env_vars, **job.env_vars},
            }
        )
    return work_list


def cmd2str(cmd: Union[str, Sequence[str]]) -> str:
    if isinstance(cmd, str):
        return cmd
    else:
        return " ".join(list(map(shlex.quote, cmd)))


def create_array_wrapper_script(
    cmd: Union[str, Sequence[str]], work_list: List[_WorkItem], task_offset: int
) -> str:
    """Create a wrapper script for running parameter sweep."""
    env_vars = _create_env_vars([work.get("env_vars", {}) for work in work_list])
    args = _create_args([work.get("args", []) for work in work_list])
    script_args = {
        "shebang": _DEFAULT_ARRAY_WRAPPER_SHEBANG,
        "task_offset": task_offset,
        "env_vars": env_vars,
        "args": args,
        "cmd": cmd2str(cmd),
    }
    script = _ARRAY_WRAPPER_TEMPLATE % script_args
    return script


def create_job_script(
    cmd: str,
    header: str,
    setup: str = "",
    shebang="#!/usr/bin/bash -l",
) -> str:
    return """\
%(shebang)s
%(header)s

%(setup)s
%(cmd)s

""" % {
        "cmd": cmd2str(cmd),
        "shebang": shebang,
        "setup": setup,
        "header": header,
    }


def validate_same_job_configuration(jobs: List[xm.Job]):
    executable = jobs[0].executable
    executor = jobs[0].executor
    for job in jobs[1:]:
        if job.executable != executable:
            raise ValueError("All jobs must have the same executable.")
        if job.executor != executor:
            raise ValueError("All jobs must have the same executor.")


def get_singulation_options(
    singularity_options: Optional[executors.SingularityOptions], use_gpu: bool
) -> List[str]:
    options = singularity_options or executors.SingularityOptions()
    result = []

    if options.bind is not None:
        for host, container in options.bind.items():
            result.extend(["--bind", f"{host}:{container}"])

    result.extend(list(options.extra_options))

    if use_gpu:
        result.append("--nv")
    return result



import functools
import os
from collections import UserDict
from typing import Any, Dict, Optional

import appdirs
import tomlkit
from absl import flags

LXM_CONFIG = flags.DEFINE_string(
    "lxm_config", os.environ.get("LXM_CONFIG"), "Path to lxm configuration."
)


class Config(UserDict):
    @classmethod
    def from_file(cls, path: str) -> "Config":
        with open(path, "rt") as f:
            config_dict = tomlkit.loads(f.read())
        return cls(config_dict)

    @classmethod
    def from_string(cls, content: str) -> "Config":
        config_dict = tomlkit.loads(content)
        return cls(config_dict)

    def project(self) -> Optional[str]:
        project = os.environ.get("LXM_PROJECT", None)
        if project is not None:
            return project
        return self.data.get("project", None)

    def local_config(self) -> Dict[str, Any]:
        return self.data["local"]

    def default_cluster(self) -> str:
        cluster = os.environ.get("LXM_CLUSTER", None)
        if cluster is None:
            cluster = self.data["clusters"][0]["name"]
        return cluster

    def cluster_config(self, location: Optional[str] = None) -> Dict[str, Any]:
        location = location or self.default_cluster()
        clusters = {cluster["name"]: cluster for cluster in self.data["clusters"]}
        if location not in clusters:
            raise ValueError("Unknown cluster")
        cluster = clusters[location]
        return cluster


@functools.lru_cache()
def default() -> Config:
    # The path may be relative, especially if it comes from sys.argv[0].
    if LXM_CONFIG.value is not None:
        return Config.from_file(LXM_CONFIG.value)
    # Use file from environment variable
    lxm_config_path = os.environ.get("LXM_CONFIG", None)
    if lxm_config_path is not None:
        return Config.from_file(lxm_config_path)
    cwd_path = os.path.join(os.getcwd(), "lxm.toml")
    if os.path.exists(cwd_path):
        return Config.from_file(cwd_path)
    user_config_path = os.path.join(appdirs.user_config_dir("lxm3"), "config.toml")
    if os.path.exists(user_config_path):
        return Config.from_file(user_config_path)
    else:
        raise ValueError("Unable to load Config")



# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""An utility to batch Packageables together and build them in one go."""

import asyncio
import concurrent.futures as concurrent_futures
import threading
from typing import Awaitable, Callable, Sequence, TypeVar, Union

from lxm3._vendor.xmanager.xm import job_blocks


class PackageHasNotBeenCalledError(RuntimeError):
  """Access to package_async() awaitable prior to calling .package()."""


Awaited = TypeVar('Awaited')


class PicklableAwaitableImpl:
  """Awaitable type with known value which can be pickled."""

  def __init__(
      self,
      get_future: Callable[
          [], Union[asyncio.Future[Awaited], concurrent_futures.Future[Awaited]]
      ],
  ):
    self._get_future = get_future

  def __await__(self):
    return asyncio.wrap_future(self._get_future()).__await__()

  def __reduce__(self):
    return _return_awaited, (self._get_future().result(),)


def _return_awaited(
    awaited: Awaited,
) -> Awaitable[Awaited]:
  """Returns a picklable awaitable for an already known value."""

  def get_future() -> asyncio.Future[Awaited]:
    future = asyncio.Future()
    future.set_result(awaited)
    return future

  return PicklableAwaitableImpl(get_future)


class AsyncPackager:
  """An utility to batch Packageables together and build them in one go.

  Attributes:
    _lock: A Lock() object used to make the class threadsafe.
    _package_batch: A function which packages a batch of Packageables.
    _packageables: Packageables queued to be packaged.
    _futures: Corresponding futures where packaging results should be written.
  """

  def __init__(
      self,
      package_batch: Callable[
          [Sequence[job_blocks.Packageable]], Sequence[job_blocks.Executable]
      ],
  ) -> None:
    """Creates the async packager.

    Args:
      package_batch: A function which packages a batch of Packageables.
    """
    super().__init__()
    self._lock = threading.Lock()
    self._package_batch = package_batch
    self._packageables = []
    self._futures = []

  def add(
      self, packageable: job_blocks.Packageable
  ) -> Awaitable[job_blocks.Executable]:
    """Adds new packageable to the batch."""
    with self._lock:
      future = concurrent_futures.Future()
      self._packageables.append(packageable)
      self._futures.append(future)

    def check_is_packaged() -> None:
      with self._lock:
        if packageable in self._packageables:
          raise PackageHasNotBeenCalledError(
              '.package() must be called before awaiting on the packaging '
              'result'
          )

    def get_future() -> concurrent_futures.Future[job_blocks.Executable]:
      check_is_packaged()
      return future

    return PicklableAwaitableImpl(get_future)

  def package(
      self, extra_packageables: Sequence[job_blocks.Packageable] = ()
  ) -> Sequence[job_blocks.Executable]:
    """Triggers the packaging of previously added packageables.

    Args:
      extra_packageables: An explicit sequence of extra packageables items to
        package synchronously.

    Returns:
      The list of executables corresponding to `extra_packageables`.
    """
    with self._lock:
      packageables = self._packageables + list(extra_packageables)
      futures = self._futures
      self._packageables = []
      self._futures = []

    if not packageables:
      return []

    try:
      executables = self._package_batch(packageables)
      for executable, future in zip(executables, futures):
        future.set_result(executable)
      return executables[len(futures) :]
    except Exception as e:
      for future in futures:
        future.set_exception(e)
      raise



import datetime
import os
from typing import List, Optional

from absl import logging

from lxm3 import xm
from lxm3.clusters import gridengine
from lxm3.xm_cluster import config as config_lib
from lxm3.xm_cluster import executables
from lxm3.xm_cluster import executors
from lxm3.xm_cluster import requirements as cluster_requirements
from lxm3.xm_cluster.console import console
from lxm3.xm_cluster.execution import artifacts
from lxm3.xm_cluster.execution import common
from lxm3.xm_cluster.execution import job_script

_TASK_OFFSET = 1
_JOB_SCRIPT_SHEBANG = "#!/usr/bin/env bash"
_TASK_ID_VAR_NAME = "SGE_TASK_ID"


def _format_time(duration_seconds: int) -> str:
    hours, remainder = divmod(duration_seconds, 3600)
    minutes, seconds = divmod(remainder, 60)
    return "{:02}:{:02}:{:02}".format(int(hours), int(minutes), int(seconds))


def _generate_header_from_executor(
    job_name: str,
    executor: executors.GridEngine,
    num_array_tasks: Optional[int],
    job_script_dir: str,
) -> str:
    header = []

    header.append(f"#$ -N {job_name}")

    if executor.requirements.task_requirements:
        logging.warning(
            "Task requirements in GridEngine will be ignored."
            "Please use GridEngine.resources for customizing resources."
        )

    for resource, value in executor.resources.items():
        if value:
            header.append(f"#$ -l {resource}={value}")

    for pe_name, value in executor.parallel_environments.items():
        if value > 0:
            header.append(f"#$ -pe {pe_name} {value}")

    if executor.walltime is not None:
        duration_secs = int(executor.walltime.total_seconds())
        header.append(f"#$ -l h_rt={_format_time(duration_secs)}")

    if executor.queue:
        header.append(f"#$ -q {executor.queue}")

    reserved = executor.reserved
    if reserved is None:
        if (
            executor.parallel_environments
            or cluster_requirements.ResourceType.GPU
            in executor.requirements.task_requirements
            or "gpu" in executor.resources
        ):
            reserved = True

    if reserved:
        header.append("#$ -R y")

    log_directory = executor.log_directory or os.path.join(job_script_dir, "logs")
    if num_array_tasks is not None:
        stdout = os.path.join(log_directory, "$JOB_NAME.o$JOB_ID.$TASK_ID")
        stderr = os.path.join(log_directory, "$JOB_NAME.e$JOB_ID.$TASK_ID")
    else:
        stdout = os.path.join(log_directory, "$JOB_NAME.o$JOB_ID")
        stderr = os.path.join(log_directory, "$JOB_NAME.e$JOB_ID")

    header.append(f"#$ -o {stdout}")
    header.append(f"#$ -e {stderr}")
    if executor.merge_output:
        header.append("#$ -j y")

    if num_array_tasks is not None:
        header.append(f"#$ -t 1-{num_array_tasks}")
        if executor.max_parallel_tasks:
            header.append(f"#$ -tc {executor.max_parallel_tasks}")

    # Use current working directory
    header.append("#$ -cwd")

    header.append(f"#$ -S {executor.shell}")

    if executor.project:
        header.append(f"#$ -P {executor.project}")

    if executor.account:
        header.append(f"#$ -A {executor.account}")

    # Skip requested header directives
    header = list(
        filter(
            lambda line: not any(skip in line for skip in executor.skip_directives),
            header,
        )
    )

    for line in executor.extra_directives:
        if not line.startswith("#$"):
            line = "#$ " + line
        header.append(line)

    return "\n".join(header)


def _is_gpu_requested(executor: executors.GridEngine) -> bool:
    return (
        "gpu" in executor.parallel_environments
        or cluster_requirements.ResourceType.GPU
        in executor.requirements.task_requirements
        or "gpu" in executor.resources
    )


def _create_job_header(
    executor: executors.GridEngine,
    jobs: List[xm.Job],
    job_script_dir: str,
    job_name: str,
) -> str:
    num_array_tasks = len(jobs) if len(jobs) > 1 else None
    job_header = _generate_header_from_executor(
        job_name, executor, num_array_tasks, job_script_dir
    )
    return job_header


def _get_setup_cmds(
    executable: executables.Command,
    executor: executors.GridEngine,
) -> str:
    cmds = ["echo >&2 INFO[$(basename $0)]: Running on host $(hostname)"]

    for module in executor.modules:
        cmds.append(f"module load {module}")
    if _is_gpu_requested(executor):
        cmds.append(
            "echo >&2 INFO[$(basename $0)]: CUDA_VISIBLE_DEVICES=$CUDA_VISIBLE_DEVICES"
        )
        cmds.append("nvidia-smi")

    if executable.singularity_image is not None:
        cmds.append(
            "echo >&2 INFO[$(basename $0)]: Singularity version: $(singularity --version)"
        )
    return "\n".join(cmds)


def deploy_job_resources(
    artifact: artifacts.Artifact,
    jobs: List[xm.Job],
    version: Optional[str] = None,
) -> str:
    version = version or datetime.datetime.now().strftime("%Y%m%d.%H%M%S")
    executable = jobs[0].executable
    executor = jobs[0].executor

    assert isinstance(executor, executors.GridEngine)
    assert isinstance(executable, executables.Command)
    job_script.validate_same_job_configuration(jobs)

    job_name = f"job-{version}"

    job_script_dir = artifact.job_path(job_name)

    setup = _get_setup_cmds(executable, executor)
    header = _create_job_header(executor, jobs, job_script_dir, job_name)

    return common.create_array_job(
        artifact=artifact,
        executable=executable,
        singularity_image=executable.singularity_image,
        singularity_options=executor.singularity_options,
        jobs=jobs,
        use_gpu=_is_gpu_requested(executor),
        version=version,
        job_script_shebang=_JOB_SCRIPT_SHEBANG,
        task_offset=_TASK_OFFSET,
        task_id_var_name=_TASK_ID_VAR_NAME,
        setup=setup,
        header=header,
    )


class GridEngineHandle:
    def __init__(self, job_id: str) -> None:
        self.job_id = job_id

    async def wait(self):
        raise NotImplementedError()

    async def monitor(self):
        raise NotImplementedError()


async def launch(
    config: config_lib.Config, jobs: List[xm.Job]
) -> List[GridEngineHandle]:
    if len(jobs) < 1:
        return []

    storage_root, hostname, user, connect_kwargs = common.get_cluster_settings(
        config, jobs
    )

    artifact = artifacts.create_artifact_store(
        storage_root,
        hostname=hostname,
        user=user,
        project=config.project(),
        connect_kwargs=connect_kwargs,
    )

    client = gridengine.Client(hostname, user)
    console.log(f"Launching {len(jobs)} jobs on {hostname}")
    job_script_path = deploy_job_resources(artifact, jobs)

    console.log(f"Launch with command:\n  qsub {job_script_path}")
    group = client.launch(job_script_path)
    common.write_job_id(artifact, job_script_path, group.group(0))
    console.log(f"Successfully launched job {group.group(0)}")
    job_ids = gridengine.split_job_ids(group)

    return [GridEngineHandle(job_id) for job_id in job_ids]



import abc
import datetime
import os
from typing import Any, Mapping, Optional

import fsspec
import rich.progress
import rich.syntax
from fsspec.implementations.sftp import SFTPFileSystem

from lxm3.xm_cluster.console import console


class Artifact(abc.ABC):
    def __init__(
        self,
        filesystem: fsspec.AbstractFileSystem,
        staging_directory: str,
        project: Optional[str] = None,
    ):
        if project:
            self._storage_root = os.path.join(staging_directory, "projects", project)
        else:
            self._storage_root = staging_directory
        self._fs = filesystem
        self._initialize()

    def _initialize(self):
        pass

    def job_path(self, job_name: str):
        return os.path.join(self._storage_root, "jobs", job_name)

    def job_script_path(self, job_name: str):
        return os.path.join(self.job_path(job_name), "job.sh")

    def job_array_wrapper_path(self, job_name: str):
        return os.path.join(self.job_path(job_name), "array_wrapper.sh")

    def singularity_image_path(self, image_name: str):
        return os.path.join(self._storage_root, "containers", image_name)

    def archive_path(self, resource_uri: str):
        return os.path.join(
            self._storage_root, "archives", os.path.basename(resource_uri)
        )

    def _should_update(self, src: str, dst: str) -> bool:
        if not self._fs.exists(dst):
            return True

        local_stat = os.stat(src)
        local_mtime = datetime.datetime.utcfromtimestamp(
            local_stat.st_mtime
        ).timestamp()
        storage_stat = self._fs.info(dst)
        storage_mtime = storage_stat["mtime"]

        if isinstance(storage_mtime, datetime.datetime):
            storage_mtime = storage_mtime.timestamp()

        if local_stat.st_size != storage_stat["size"]:
            return True

        if int(local_mtime) > int(storage_mtime):
            return True

        return False

    def _put_content(self, dst, content):
        self._fs.makedirs(os.path.dirname(dst), exist_ok=True)

        with self._fs.open(dst, "wt") as f:
            f.write(content)

    def _put_file(self, local_filename, dst):
        self._fs.makedirs(os.path.dirname(dst), exist_ok=True)
        self._fs.put(local_filename, dst)

    def deploy_job_scripts(self, job_name, job_script, array_wrapper=None):
        job_path = self.job_path(job_name)
        job_log_path = os.path.join(job_path, "logs")

        self._fs.makedirs(job_path, exist_ok=True)
        self._fs.makedirs(job_log_path, exist_ok=True)
        self._put_content(self.job_script_path(job_name), job_script)
        if array_wrapper is not None:
            self._put_content(self.job_array_wrapper_path(job_name), array_wrapper)
        console.log(f"Created job script {self.job_script_path(job_name)}")

    def deploy_singularity_container(self, singularity_image):
        image_name = os.path.basename(singularity_image)
        deploy_container_path = self.singularity_image_path(image_name)
        should_update = self._should_update(singularity_image, deploy_container_path)
        if should_update:
            self._fs.makedirs(os.path.dirname(deploy_container_path), exist_ok=True)
            self._put_file(singularity_image, deploy_container_path)
            console.log(f"Deployed Singularity container to {deploy_container_path}")
        else:
            console.log(f"Container {deploy_container_path} exists, skip upload.")
        return deploy_container_path

    def deploy_resource_archive(self, resource_uri):
        deploy_archive_path = self.archive_path(resource_uri)

        should_update = self._should_update(resource_uri, deploy_archive_path)
        if should_update:
            self._put_file(resource_uri, deploy_archive_path)
            console.log(f"Deployed archive to {deploy_archive_path}")
        else:
            console.log(f"Archive {deploy_archive_path} exists, skipping upload.")
        return deploy_archive_path


class LocalArtifact(Artifact):
    def __init__(self, staging_directory: str, project=None):
        staging_directory = os.path.abspath(os.path.expanduser(staging_directory))
        super().__init__(fsspec.filesystem("file"), staging_directory, project)

    def _initialize(self):
        if not os.path.exists(self._storage_root):
            self._fs.makedirs(self._storage_root)
            # Create a .gitignore file to prevent git from tracking the directory
            self._fs.write_text(os.path.join(self._storage_root, ".gitignore"), "*\n")

    def deploy_singularity_container(self, singularity_image):
        image_name = os.path.basename(singularity_image)
        deploy_container_path = self.singularity_image_path(image_name)
        should_update = self._should_update(singularity_image, deploy_container_path)
        if should_update:
            self._fs.makedirs(os.path.dirname(deploy_container_path), exist_ok=True)
            if os.path.exists(deploy_container_path):
                os.unlink(deploy_container_path)
            self._put_file(singularity_image, deploy_container_path)
            console.log(f"Deployed Singularity container to {deploy_container_path}")
        else:
            console.log(f"Container {deploy_container_path} exists, skip upload.")
        return deploy_container_path


class RemoteArtifact(Artifact):
    _fs: SFTPFileSystem

    def __init__(
        self,
        hostname: str,
        user: Optional[str],
        staging_directory: str,
        project: Optional[str] = None,
        connect_kwargs: Optional[Mapping[str, Any]] = None,
    ):
        if connect_kwargs is None:
            connect_kwargs = {}
        fs = fsspec.filesystem("sftp", host=hostname, username=user, **connect_kwargs)
        # Normalize the storage root to an absolute path.
        self._host = hostname
        self._user = user
        if not os.path.isabs(staging_directory):
            staging_directory = fs.ftp.normalize(staging_directory)
        super().__init__(fs, staging_directory, project)

    def deploy_singularity_container(self, singularity_image):
        image_name = os.path.basename(singularity_image)
        deploy_container_path = self.singularity_image_path(image_name)
        should_update = self._should_update(singularity_image, deploy_container_path)
        if should_update:
            with rich.progress.Progress(
                rich.progress.TextColumn("[progress.description]{task.description}"),
                rich.progress.BarColumn(),
                rich.progress.TaskProgressColumn(),
                rich.progress.TimeRemainingColumn(),
                rich.progress.TransferSpeedColumn(),
                console=console,
            ) as progress:
                self._fs.makedirs(os.path.dirname(deploy_container_path), exist_ok=True)

                task = progress.add_task(
                    f"Uploading {os.path.basename(singularity_image)}"
                )

                def callback(transferred_bytes: int, total_bytes: int):
                    progress.update(
                        task, completed=transferred_bytes, total=total_bytes
                    )

                self._fs.ftp.put(
                    singularity_image,
                    deploy_container_path,
                    callback=callback,
                    confirm=True,
                )
                progress.update(task, description="Done!")

        else:
            console.log(f"Container {deploy_container_path} exists, skip upload.")
        return deploy_container_path


def create_artifact_store(
    storage_root: str,
    hostname: Optional[str] = None,
    user: Optional[str] = None,
    project: Optional[str] = None,
    connect_kwargs=None,
):
    if hostname is None:
        return LocalArtifact(storage_root, project=project)
    else:
        return RemoteArtifact(
            hostname,
            user,
            storage_root,
            project=project,
            connect_kwargs=connect_kwargs,
        )



import dataclasses
import inspect
from typing import Any, Dict, Iterator, List, Optional, Type

import fiddle
from fiddle import config
from fiddle import daglish
from fiddle.experimental import serialization


def _has_nested_builder(value: Any, state=None) -> bool:
    state = state or daglish.MemoizedTraversal.begin(_has_nested_builder, value)
    return isinstance(value, config.Buildable) or (
        state.is_traversable(value) and any(state.flattened_map_children(value).values)
    )


def _path_str(path: daglish.Path) -> str:
    """Formats path in a way customized to this file.

    In the future, we may wish to consider a format that is readable for printing
    more arbitrary collections.

    Args:
      path: A path, which typically starts with an attribute.

    Returns:
      String representation of the path.
    """
    path_str = daglish.path_str(path)
    return path_str[1:] if path and isinstance(path[0], daglish.Attr) else path_str


@dataclasses.dataclass(frozen=True)
class _LeafSetting:
    """Represents a leaf configuration setting."""

    path: daglish.Path
    annotation: Optional[Type[Any]]
    value: Any


def as_flattened_dict(
    cfg: config.Buildable, include_buildable: bool = True
) -> Dict[str, Any]:
    """Returns a flattened dict of cfg's paths (dot syntax) and values.

    Default values and tags won't be included in the flattened dict.

    Args:
      cfg: A buildable to generate a flattened dict representation for.

    Returns: A flattened Dict representation of `cfg`.
    """

    def dict_generate(value, state=None) -> Iterator[_LeafSetting]:
        state = state or daglish.BasicTraversal.begin(dict_generate, value)

        if not _has_nested_builder(value):
            yield _LeafSetting(state.current_path, None, value)
        else:
            # value must be a Buildable or a traversable containing a Buildable.
            assert state.is_traversable(value)
            if isinstance(value, config.Buildable) and include_buildable:
                module = inspect.getmodule(value.__fn_or_cls__).__name__  # type: ignore
                name = value.__fn_or_cls__.__qualname__
                ctor_name = f"{module}:{name}"
                yield _LeafSetting(
                    state.current_path + (daglish.Attr("__ctor__"),), None, ctor_name
                )
            for sub_result in state.flattened_map_children(value).values:
                yield from sub_result

    args_dict = {}
    for leaf in dict_generate(cfg):
        args_dict[_path_str(leaf.path)] = leaf.value

    return args_dict


def basic_callable(a: int = 1):
    pass


@dataclasses.dataclass
class ExperimentConfig:
    fn: Any
    seed: int
    name: str
    test: bool
    nest: Dict
    list_: List
    unset: int = 1


if __name__ == "__main__":
    import wandb

    wandb.init(project="lxm3", mode="disabled")

    cfg = fiddle.Config(
        ExperimentConfig,
        fn=fiddle.Partial(basic_callable, a=2),
        seed=42,
        name="test",
        test=True,
        nest={"a": 1, "b": 2},
        list_=[1, 2, 3],
    )

    # artifact.add_file("config.json", serialization.dump_json(cfg))
    dumped = serialization.dump_json(cfg)
    artifact = wandb.Artifact(name="config", type="fiddle_config")
    with artifact.new_file("config.json", "wt") as f:
        f.write(dumped)
    wandb.log_artifact(artifact)
    flat = as_flattened_dict(cfg)
    print(flat)



import os
import re
from typing import List, NamedTuple, Union

import attr

from lxm3 import xm
from lxm3.xm import executables
from lxm3.xm import job_blocks
from lxm3.xm import utils


def name_from_path(path: str) -> str:
    """Returns a safe to use executable name based on a filesystem path."""
    return re.sub("\\W", "_", os.path.basename(path.rstrip(os.sep)))


class Fileset(executables.BinaryDependency):
    """Additional files to package."""

    def __init__(self, files=None):
        """Create a Fileset.

        Args:
            files: A mapping from local_path -> fileset path.
        """
        files = files or {}
        self._files = {}
        for src, dst in files.items():
            self.add_path(src, dst)

    def add_path(self, src: str, dst: str):
        """Add a file to the fileset.

        Args:
            src: Path to the file on the local filesystem.
            dst: Path to the file in the package.
        """
        self._files[dst] = src

    def get_path(self, name: str, executor_spec: xm.ExecutableSpec):
        """Resolve the path to a file in fileset."""
        del executor_spec
        # Currently no-op
        return name

    @property
    def files(self):
        return [(src, dst) for dst, src in self._files.items()]


class ModuleName(NamedTuple):
    """Name of python module to execute when entering this project."""

    module_name: str


class CommandList(NamedTuple):
    """List of commands to execute when entering this project."""

    commands: List[str]


@attr.s(auto_attribs=True)
class PythonPackage(job_blocks.ExecutableSpec):
    """Python package describes an executable that can be packaged by
    ``pip install``.

    :obj:`PythonPackage` describes a python distribution that can be
    installed by ``pip install``. This is the recommended way to
    package python projects. For an introduction to python packaging,
    see
    `Packaging Python Projects <https://packaging.python.org/en/latest/tutorials/packaging-projects/>`_.

    Note:
        lxm3 uses ``pip install`` to install the package into a temporary
        directory that will be subsequently packaged into a zip archive
        that will be deployed and unzipped when jobs run on the cluster.
        However, this should be considered an implementation detail.

        lxm3 uses ``pip install --no-deps`` by default to install the package.
        This means that lxm3 **will not** install the dependencies required by
        the project. The reason is that typical ML dependencies are
        large. (e.g. TensorFlow is around 400MB)
        Therefore, you should install your dependencies
        first on the cluster (better, if you are using singularity,
        install them in the container).

        The root directory of the deployed archive will be added to
        PYTHONPATH, so you can import your package as usual. However,
        if that interferes with package imports, please open an issue
        and let us know.

    Attributes:
        entrypoint: Entrypoint for the built executable.

            Currently, only :obj:`ModuleName` and :obj:`CommandList`
            are supported.

            For :obj:`ModuleName`, this corresponds to running
            ``python3 -m <module_name>``.

            For :obj:`CommandList`, this corresponds to running a
            shell script including the commands.

        path: Path to the python project.
            This should be the path to a directory ``path`` that can
            be installed as a python package by ``pip install
            <path>``. If it's a relative path, this will be resolved
            relative to the launcher's working directory.

        resources: List of resources to package. Currently, only
            :obj:`Fileset` is supported.

        extra_packages: List of paths to additional packages that will
            be install by ``pip``.

            :attr:`extra_packages` is useful if you also
            want to include a dependency that you are also working on
            locally and you want to use your development version which
            are not installed in the runtime environment.

        pip_args: Additional options that will be passed to ``pip install``.
            Defaults to ``--no-deps --no-compile``.
    """

    entrypoint: Union[CommandList, ModuleName]
    path: str = attr.ib(converter=utils.resolve_path_relative_to_launcher, default=".")
    resources: List[Fileset] = attr.ib(converter=list, default=attr.Factory(list))
    extra_packages: List[str] = attr.ib(converter=list, default=attr.Factory(list))
    pip_args: List[str] = attr.ib(
        converter=list, default=attr.Factory(lambda: ["--no-deps", "--no-compile"])
    )

    @property
    def name(self) -> str:
        return name_from_path(self.path)


@attr.s(auto_attribs=True)
class UniversalPackage(job_blocks.ExecutableSpec):
    """Universal package describes a package that can be built by a custom build script.

    Compared to `PythonPackage`, UniversalPackage is more flexible,
    as it can be used for any language and build system not supported natived by LXM3.
    However, it requires more work to set up.

    Args:
        entrypoint: Entrypoint for the built executable.
        build_script: Path to the build script. If it's a relative path, this will be
            resolved relative to `path`.
            The build script should be an executable that can be used to produce a
            directory containing files that will be packaged into a zip archive.
            During packaging. The build script put the files into the directory
            specified by the ``BUILDDIR`` environment variable.
        build_args: Additional arguments that will be passed to the build script.
            path: Path to the project.
        path: Path to the project. If it's a relative path, this will be resolved
            relative to the launcher's working directory.

    Examples:
        See ``examples/universal_package`` for an example.

    Raises:
        ValueError: If the build script is not executable.

    """

    entrypoint: List[str]
    build_script: str
    build_args: List[str] = attr.ib(converter=list, default=attr.Factory(list))
    path: str = attr.ib(converter=utils.resolve_path_relative_to_launcher, default=".")

    @property
    def name(self) -> str:
        return name_from_path(self.path)

    def __attrs_post_init__(self):
        self.build_script = os.path.join(self.path, self.build_script)
        if not os.access(self.build_script, os.X_OK):
            raise ValueError(
                "Build script is not executable"
                f"You may need to run `chmod +x {self.build_script}` to make it executable."
            )


@attr.s(auto_attribs=True)
class SingularityContainer(job_blocks.ExecutableSpec):
    """An executable that can be executed in a Singularity container.

    Attributes:
        entrypoint: Another ExcutableSpec.
        image_path: Path to a Singularity image.
            If it's a relative path, this will be resolved relative
            to the launcher's working directory.
    """

    entrypoint: Union[UniversalPackage, PythonPackage]
    image_path: str

    @property
    def name(self) -> str:
        return self.entrypoint.name

    def __attrs_post_init__(self):
        image_path = self.image_path
        if not os.path.isabs(self.image_path):
            image_path = utils.resolve_path_relative_to_launcher(image_path)
        if not os.path.exists(image_path):
            raise ValueError(
                f"Unable to find Singularity image at {image_path}"
                "If you use a relative path, it should be relative to the "
                "launcher's directory."
            )
        self.image_path = image_path



import os

from absl import app
from absl import flags

_SEED = flags.DEFINE_integer("seed", 0, "Random seed")


def main(_):
    print("--seed", _SEED.value, "TASK", os.environ["TASK"])


if __name__ == "__main__":
    app.run(main)



# type: ignore

import IPython
from absl import app
from traitlets.config import Config

from lxm3.clusters import gridengine
from lxm3.xm_cluster import config as config_lib
from lxm3.xm_cluster.console import console


def main(_):
    config = config_lib.default()
    location = config.default_cluster()
    cluster_config = config.cluster_config(location)
    hostname = cluster_config["server"]
    user = cluster_config["user"]

    console.log(f"Creating a client to cluster {user}@{hostname} ...")

    c = Config()
    c.InteractiveShellApp.exec_lines = [
        "%load_ext rich",
    ]

    client = gridengine.Client(hostname, user)
    IPython.start_ipython(argv=[], config=c, user_ns={"client": client})


if __name__ == "__main__":
    app.run(main)



# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Abstract API specification for XManager implementations.

Each implementation of the XManager API should override the abstract methods.
Users are normally expected to have the following pair of imports:

```
from lxm3._vendor.xmanager import xm
from lxm3._vendor.xmanager import xm_foo
```
"""

import abc
import asyncio
from concurrent import futures
import contextvars
import enum
import getpass
import inspect
import queue
import threading
import traceback
from typing import Any, Awaitable, Callable, Collection, Coroutine, Dict, Generator, List, Mapping, Optional, Sequence, overload

from absl import logging
import attr
from lxm3._vendor.xmanager.xm import async_packager
from lxm3._vendor.xmanager.xm import id_predictor
from lxm3._vendor.xmanager.xm import job_blocks
from lxm3._vendor.xmanager.xm import job_operators
from lxm3._vendor.xmanager.xm import metadata_context
from lxm3._vendor.xmanager.xm import pattern_matching

# ContextVars holding the current experiment (when within an Experiment context)
# and current experiment unit (when inside a JobGenerator).
_current_experiment: contextvars.ContextVar['Experiment'] = (
    contextvars.ContextVar('_xm_current_experiment')
)
_current_experiment_unit: contextvars.ContextVar['ExperimentUnit'] = (
    contextvars.ContextVar('_xm_current_experiment_unit')
)


def _check_if_unsupported_args_are_present(
    args: Mapping[str, Any], supported_args: Collection[str], job_type: str
) -> None:
  supported_args = set(supported_args)
  unsupported_args = set(args.keys()) - supported_args
  if unsupported_args:
    raise ValueError(
        f'Arguments {unsupported_args!r} are not supported by {job_type}. Only '
        f'{supported_args!r} are allowed.'
    )


def _apply_args_to_job(job: job_blocks.Job, args: Mapping[str, Any]) -> None:
  """Overrides job properties."""
  _check_if_unsupported_args_are_present(args, ('args', 'env_vars'), 'xm.Job')

  if 'args' in args:
    job.args = job_blocks.merge_args(job.args, args['args'])
  if 'env_vars' in args:
    job.env_vars = job.env_vars.copy()
    job.env_vars.update(args['env_vars'])


def _apply_args_to_job_group(
    job_group: job_blocks.JobGroup, args: Mapping[str, Any]
) -> None:
  """Recursively overrides job group properties."""
  if args:
    _check_if_unsupported_args_are_present(
        args, job_group.jobs.keys(), 'xm.JobGroup'
    )
    for key, job in job_group.jobs.items():
      _apply_args(job, args.get(key, {}))


_apply_args = pattern_matching.match(
    _apply_args_to_job,
    _apply_args_to_job_group,
    pattern_matching.Case(
        [job_blocks.JobGeneratorType, Any], lambda other, args: None
    ),
    pattern_matching.Case(
        [job_blocks.JobConfig, Any], lambda other, args: None
    ),
)


class ExperimentUnitStatus(abc.ABC):
  """The status of an experiment unit."""

  @property
  @abc.abstractmethod
  def is_active(self) -> bool:
    """Returns whether the unit is not in terminal state.

    It may be actively running or queued. The unit may produce more results.
    If the unit is stopped by a user it will be neither active, completed
    nor failed.
    """
    raise NotImplementedError

  @property
  @abc.abstractmethod
  def is_completed(self) -> bool:
    """Returns whether the unit has completed without failures.

    This is a terminal state. The unit has produced all the intended results.
    But it still may be restarted by an explicit request.
    """
    raise NotImplementedError

  @property
  @abc.abstractmethod
  def is_failed(self) -> bool:
    """Returns whether the unit has failed.

    This is a terminal state. Experiment unit will enter this state on any
    fatal failure, such as process exiting with non-zero code, cloud rejecting
    to schedule/queue the job or exceptions in JobGenerator. The unit will stay
    in this state unless explicitly restarted.
    Intermediate failures do not result in this state.
    """
    raise NotImplementedError

  @property
  @abc.abstractmethod
  def message(self) -> str:
    """An optional human-readable message providing context for the status.

    This may take the form of explaining why the work unit is in this state,
    or any potentially transient errors the work unit may be experiencing.
    """
    raise NotImplementedError


class ExperimentUnitError(RuntimeError):
  """Experiment unit could not be completed.

  Attrs:
    work_unit: The work unit in which the error occured, if available.
  """

  work_unit: Optional['WorkUnit'] = None

  def __init__(self, message: Any, *, work_unit: Optional['WorkUnit'] = None):
    super().__init__(message)
    self.work_unit = work_unit


class ExperimentUnitFailedError(ExperimentUnitError):
  """A job running in an experiment unit has failed."""


class ExperimentUnitNotCompletedError(ExperimentUnitError):
  """Experiment unit is neither running nor completed.

  For example it may be stopped by a user.
  """


class NotFoundError(KeyError):
  """Experiment/Work Unit/etc. has not been found."""


def _work_unit_arguments(
    job: job_blocks.JobType,
    args: Optional[Mapping[str, Any]],
) -> Mapping[str, Any]:
  """Constructs work unit arguments to display them in various UIs.

  If users pass `args` to the `.add` method explicitly, we assume `args` to be
  the sought work unit arguments. If `args` are not passed to `.add`, we deduce
  work unit arguments implicitly from the `job`s' `args` and `env_vars`.

  Args:
    job: A job to run inside a work unit.
    args: Explicitly specified arguments (could be empty).

  Returns:
    Depending on the type of the `job` given, can be one of the following:

      - if it's an instance of `Job`, we return `{'args': job.args, 'env_vars':
        job.env_vars}` with empty values omitted;
      - if it's an instance of `JobGroup`, we recursively unwind the group while
        populating corresponding nested dictionaries until we reach standalone
        `Job`s;
      - if it's a job generator, we return `{}`.
  """
  if args is not None:
    # In order to give users control on what is shown as work unit arguments we
    # don't alter them if a value is given.
    return args

  def deduce_args_for_job(job: job_blocks.Job) -> Dict[str, Any]:
    args = {
        'args': job.args.to_dict(kwargs_only=True),
        'env_vars': job.env_vars,
    }
    return {key: value for key, value in args.items() if value}

  def deduce_args_for_job_group(group: job_blocks.JobGroup) -> Dict[str, Any]:
    args = {}
    for job_name, job in group.jobs.items():
      job_args = deduce_args(job)
      if job_args:
        args[job_name] = job_args
    return args

  deduce_args = pattern_matching.match(
      deduce_args_for_job,
      deduce_args_for_job_group,
      pattern_matching.Case(
          [job_blocks.JobGeneratorType], lambda generator: {}
      ),
  )

  return deduce_args(job)


@attr.s(auto_attribs=True, kw_only=True)
class LaunchedJob:
  """A read-only view of a launched job.

  Launched jobs correspond to job instances that have been added to an
  Experiment Unit. If a Job does not have a unique identity, it can be added
  multiple times to an Experiment, so a Job object may correspond to multiple
  LaunchedJob objects. A LaunchedJob will have the same name as the added Job,
  but launched jobs may have different addresses and logs.

  `experiment.add(job)` will generate an Experiment Unit
  containing a corresponding launched job. `experiment.add(job_group)` will
  generate an Experiment Unit with multiple corresponding launched jobs.
  `experiment.add(generator)` will generate an Experiment Unit with as many
  launched jobs as the generator adds.

  Attributes:
    name: Name of the job corresponding to this launched job.
    address: The job's address.
    logs: A URL to this job's logs.
  """

  name: str
  address: Optional[str] = None
  logs: Optional[str] = None


class Importance(enum.Enum):
  """How important it is to schedule particular Experiment or ExperimentUnit.

  This is a hint to the scheduler. Not all schedulers take it into account
  (xm_local doesn't). And even with smart scheduler a less important work unit
  may run before a more important one e.g. if it uses a less contended resource.

  Unlike ServiceTier, importance only controls preference within a team i.e. how
  team's resources are divided between team's experiments. It has no effect on
  resource allocation between teams.
  """

  # High impact experiments. Try scheduling them even at the cost of significant
  # reduction of the overall throughput that your experiments get.
  HIGH = 'high'
  # The default importance.
  NORMAL = 'normal'
  # Prefer to schedule other experiments with higher importance, but in overall
  # try to maximize throughput.
  LOW = 'low'


@attr.s(auto_attribs=True, kw_only=True)
class ExperimentUnitRole(abc.ABC):
  """The role of an experiment unit within the experiment structure.

  Attributes:
    importance: how important it is to schedule this executable unit comparing
      to all your executable units (from all your experiments).
  """

  importance: Importance = Importance.NORMAL


class ExperimentUnit(abc.ABC):
  """ExperimentUnit is a collection of semantically associated `Job`s."""

  experiment: 'Experiment'

  def __init__(
      self,
      experiment: 'Experiment',
      create_task: Callable[[Awaitable[Any]], futures.Future[Any]],
      args: Optional[Mapping[str, Any]],
      role: ExperimentUnitRole,
  ) -> None:
    """Initializes an `ExperimentUnit` instance.

    Args:
      experiment: An experiment this unit belongs to.
      create_task: A callback to register a new asynchronous task.
      args: Arguments to this experiment unit. Most commonly used to represent
        the hyperparameter sweep trial corresponding to a work unit.
      role: The role of this unit in the experiment structure.
    """
    self.experiment = experiment
    self._create_task = create_task
    self._args = args
    self._role = role

    self._launch_tasks: List[futures.Future[Any]] = []

  @property
  def experiment_id(self) -> int:
    """Returns a unique ID assigned to the experiment."""
    return self.experiment.experiment_id

  def add(
      self,
      job: job_blocks.JobType,
      args: Optional[Mapping[str, Any]] = None,
      *,
      identity: str = '',
  ) -> Awaitable[None]:
    # pyformat: disable
    """Adds a Job / JobGroup to the experiment unit.

    Only one JobGroup can be added to an ExperimentUnit. This limitation may be
    lifted in future versions.

    Args:
      job: A job or job group to add.
      args: Keyword arguments to be passed to the job. For Job and JobGroup args
        are recursively expanded. For example,

        ```
        wu.add(
            JobGroup(agent=Job(...)),
            args={'agent': {'args': {'learning_rate': 0.1}}},
        )
        ```

        would update `args` field of a job `agent` in the group.
      identity: Optional unique job identifier. If not empty, `add` adopts an
        'add if not exists' behavior. No changes to existing job would be done.

    Returns:
      An awaitable that would be fulfilled when the job is launched.
    """
    # pyformat: enable
    job = job_operators.shallow_copy_job_type(job)
    if args is not None:
      _apply_args(job, args)
    job_operators.populate_job_names(job)

    def launch_job(job: job_blocks.Job) -> Awaitable[None]:
      _current_experiment.set(self.experiment)
      _current_experiment_unit.set(self)
      return self._launch_job_group(
          job_blocks.JobGroup(**{job.name: job}),
          _work_unit_arguments(job, self._args),
          identity,
      )

    def launch_job_group(group: job_blocks.JobGroup) -> Awaitable[None]:
      _current_experiment.set(self.experiment)
      _current_experiment_unit.set(self)
      return self._launch_job_group(
          group, _work_unit_arguments(group, self._args), identity
      )

    def launch_job_generator(
        job_generator: job_blocks.JobGeneratorType,
    ) -> Awaitable[None]:
      if not inspect.iscoroutinefunction(
          job_generator
      ) and not inspect.iscoroutinefunction(job_generator.__call__):
        raise ValueError(
            'Job generator must be an async function. Signature needs to be '
            '`async def job_generator(work_unit: xm.WorkUnit) -> None:`'
        )
      _current_experiment.set(self.experiment)
      _current_experiment_unit.set(self)
      coroutine = job_generator(self, **(args or {}))
      assert coroutine is not None
      return coroutine

    def launch_job_config(job_config: job_blocks.JobConfig) -> Awaitable[None]:
      _current_experiment.set(self.experiment)
      _current_experiment_unit.set(self)
      return self._launch_job_config(job_config, args or {}, identity)

    job_awaitable = pattern_matching.match(
        launch_job,
        launch_job_group,
        launch_job_generator,
        launch_job_config,
    )(job)
    launch_task = self._create_task(job_awaitable)
    self._launch_tasks.append(launch_task)
    return asyncio.wrap_future(launch_task)

  async def _wait_until_complete_impl(self) -> 'ExperimentUnit':
    try:
      for task in self._launch_tasks:
        await asyncio.wrap_future(task)
    except Exception as e:
      raise ExperimentUnitError('Experiment unit could not be created.') from e
    await self._wait_until_complete()
    return self

  def wait_until_complete(self) -> Coroutine[Any, Any, 'ExperimentUnit']:
    """Waits until the unit is in a final state: completed/failed/stopped.

    Raises:
      ExperimentUnitError: Exception if the unit couldn't complete.
    Returns:
      Returns self to facilitate asyncio.as_completed usage.
    """
    return self._wait_until_complete_impl()

  async def _launch_job_group(
      self,
      job_group: job_blocks.JobGroup,
      args_view: Mapping[str, Any],
      identity: str,
  ) -> None:
    """Launches a given job group as part of the unit."""
    raise NotImplementedError

  async def _launch_job_config(
      self,
      job_config: job_blocks.JobConfig,
      args_view: Mapping[str, Any],
      identity: str,
  ) -> None:
    """Launches a given job config as part of the unit."""
    raise NotImplementedError

  async def _wait_until_complete(self) -> None:
    """Waits until the unit is in a final state: completed/failed/stopped.

    Child classes need to implement this method to support awaiting units.

    Unlike wait_until_complete this method asumes that unit has been fully
    created. This method is only invoked if somebody has requested to monitor
    unit.
    """
    raise NotImplementedError

  def stop(
      self,
      *,
      mark_as_failed: bool = False,
      mark_as_completed: bool = False,
      message: Optional[str] = None,
  ) -> None:
    """Initiate the process to stop the unit from running.

    This method will synchronously make a request for the unit to stop.
    However, the method does not actually wait for the unit to be in a
    terminal state.

    Use self.wait_until_complete() after self.stop() to guarantee the unit
    is stopped.

    Args:
      mark_as_failed: Mark this unit as failed rather than stopped.
      mark_as_completed: Mark this unit as completed rather than stopped.
      message: Optional user-defined status message.
    """
    raise NotImplementedError

  def get_status(self) -> ExperimentUnitStatus:
    """Gets the status of this unit."""
    raise NotImplementedError

  @property
  @abc.abstractmethod
  def experiment_unit_name(self) -> str:
    raise NotImplementedError

  def get_full_job_name(self, job_name: str) -> str:
    """Given `Job.name` constructs its full name.

    The primary use case is addressing containers -- full names serve as
    hostnames.

    Args:
        job_name: Short name of a job.

    Returns:
        Full name of the job.
    """
    return f'{self.experiment_unit_name}_{job_name}'

  @property
  def context(self) -> metadata_context.MetadataContext:
    """Returns metadata context for a unit."""
    return metadata_context.MetadataContext(
        creator=getpass.getuser(),
        annotations=metadata_context.ContextAnnotations(),
    )

  @property
  def launched_jobs(self) -> List[LaunchedJob]:
    """Gets a representation for each individual job that was added.

    Each added Job should produce a LaunchedJob.
    Each added JobGroup will produce a LaunchedJob for each leaf Job.
    """
    raise NotImplementedError


@attr.s(auto_attribs=True, kw_only=True)
class WorkUnitRole(ExperimentUnitRole):
  """An experiment unit with this role is a work unit.

  Work units contain jobs that are often run as trials as part of an
  experiment's hyper-parameter search. The status of a work unit is used to
  determine the status of the experiment.
  """


class WorkUnitCompletedAwaitable(Coroutine):
  """Awaitable for work unit completion event.

  Usage:
    completion_events = [work_unit.wait_until_complete() for work_unit in ...]
    while completion_events:
      completed_event, completion_events = asyncio.wait(
          completion_events, return_when=asyncio.FIRST_COMPLETED)
      for event in completed_events:
        wid = event.work_unit.work_unit_id
        try:
          await event
          print(f'Wor unit {wid} completed successfully.')
        except xm.ExperimentUnitError as e:
          print(f'Wor unit {wid} failed: {e}.')
  """

  def __init__(
      self, work_unit: 'WorkUnit', awaitable: Callable[[], Any]
  ) -> None:
    self.work_unit = work_unit
    self._awaitable = awaitable
    self._wait_coro = self._wait()

  async def _wait(self) -> 'WorkUnit':
    # Coroutine must be created inside of async function to avoid
    # "coroutine ... was never awaited" runtime warning.
    await self._awaitable()
    return self.work_unit

  def __await__(self) -> Generator[Any, None, 'WorkUnit']:
    return self._wait_coro.__await__()

  def send(self, value: Any) -> Any:
    return self._wait_coro.send(value)

  def throw(self, typ, val=None, tb=None) -> Any:
    return self._wait_coro.throw(typ, val, tb)

  def close(self) -> None:
    self._wait_coro.close()


class WorkUnit(ExperimentUnit):
  """Work units are experiment units with the work unit role."""

  @property
  @abc.abstractmethod
  def work_unit_id(self) -> int:
    raise NotImplementedError

  def wait_until_complete(self) -> WorkUnitCompletedAwaitable:
    """Waits until the unit is in a final state: completed/failed/stopped.

    Raises:
      ExperimentUnitError: Exception if the unit couldn't complete.
    Returns:
      Returns self to facilitate asyncio.as_completed usage.
    """
    return WorkUnitCompletedAwaitable(self, self._wait_until_complete_impl)


@attr.s(auto_attribs=True, kw_only=True)
class AuxiliaryUnitRole(ExperimentUnitRole):
  """An experiment unit with this role is an auxiliary unit.

  Auxiliary units contain jobs that are not part of the trials of a
  hyper-parameter search. The status of an auxiliary unit is not used to
  determine the status of the experiment. e.g. Tensorboard

  Attributes:
    termination_delay_secs: How long to keep AUX unit running after experiment
      completion.
  """

  termination_delay_secs: int


class AuxiliaryUnitJob(abc.ABC):
  """A job bundled with an AuxiliaryUnitRole.

  This class allows libraries to define self-contained objects which would
  result in AUX units once added to the expetiment.
  Note that this class conforms to xm.JobGenerator interface.
  """

  role: AuxiliaryUnitRole
  _job: job_blocks.JobType

  def __init__(
      self,
      job: job_blocks.JobType,
      *,
      importance: Importance = Importance.NORMAL,
      termination_delay_secs: int,
  ) -> None:
    self.role = AuxiliaryUnitRole(
        importance=importance,
        termination_delay_secs=termination_delay_secs,
    )
    self._job = job

  async def __call__(self, aux_unit: ExperimentUnit, **kwargs):

    async def launch_generator(
        job_generator: job_blocks.JobGeneratorType,
    ) -> None:
      coroutine = job_generator(aux_unit, **kwargs)
      assert coroutine is not None
      await coroutine

    async def launch_job(job: Any) -> None:
      aux_unit.add(job, args=kwargs)

    await pattern_matching.async_match(launch_generator, launch_job)(self._job)


class Experiment(abc.ABC):
  """Experiment is the core unit of research in XManager.

  An experiment typically involves running a computation (e.g. training a model)
  in different hyperparameter configurations. Experiments are made up of work
  units, which do the computation(s) in question, and auxiliary units which
  perform other functions like TensorBoard. It can have associated metadata,
  such as title or source code commit from which it has been compiled.

  XManager API provides multiple implementations of the Experiment class which
  use different backends to start jobs. Types of supported Executables may vary
  between these implementations. Usually experiments woulbe be created by though
  xm_<foo>.create_experiment() functions.

  While experiment metadata and statuses can accessed and altered directly,
  adding work or auxiliary units requires entering experiment context:

    with xm_foo.create_experiment(...) as experiment:
      experiment.add(xm.Job(...))

  This context would spawn background event pool to support asynchronous
  operations and ensure that they all are completed on exit. It also may mark
  experiment as failed if an exception is thrown.
  """

  # An event loop in which job generators would be run.
  _event_loop: asyncio.AbstractEventLoop
  # A queue of background tasks that launch work units.
  _running_tasks: queue.Queue[futures.Future[Any]]
  # Work unit ID predictor.
  _work_unit_id_predictor: id_predictor.Predictor
  # A class variable for batching packaging requests.
  _async_packager: async_packager.AsyncPackager
  # ContextVars token when entering the context.
  _current_experiment_token: contextvars.Token
  _current_async_experiment_token: contextvars.Token

  @property
  def experiment_id(self) -> int:
    """Returns a unique ID assigned to the experiment."""
    raise NotImplementedError

  def __enter__(self):
    is_coro_context = False
    try:
      asyncio.get_running_loop()
      is_coro_context = True
    except RuntimeError:
      pass
    if is_coro_context:
      raise RuntimeError(
          'When using Experiment from a coroutine please use '
          '`async with` syntax'
      )

    self._current_experiment_token = _current_experiment.set(self)
    self._event_loop = asyncio.new_event_loop()
    asyncio.get_child_watcher().attach_loop(self._event_loop)
    self._event_loop_thread = threading.Thread(
        target=self._event_loop.run_forever, daemon=True
    )
    self._event_loop_thread.start()

    # asyncio.run_coroutine_threadsafe doesn't accept class method and wants it
    # wrapped in a function.
    async def async_enter():
      await self.__aenter__()

    asyncio.run_coroutine_threadsafe(
        async_enter(), loop=self._event_loop
    ).result()

    return self

  def _wait_for_tasks(self):
    """Waits for pending tasks to complete, raising the first error."""
    exception = None
    while not self._running_tasks.empty():
      try:
        self._running_tasks.get_nowait().result()
      except futures.CancelledError:
        # Ignore cancelled tasks.
        pass
      except Exception as e:  # pylint: disable=broad-except
        # Allow remaining tasks to complete before raising the first exception.
        if not exception:
          exception = e
    if exception:
      raise exception

  def __exit__(self, exc_type, exc_value, traceback):  # pylint:disable=redefined-outer-name
    _current_experiment.reset(self._current_experiment_token)
    self._wait_for_tasks()
    self._event_loop.call_soon_threadsafe(self._event_loop.stop)
    self._event_loop_thread.join()

  async def __aenter__(self):
    self._current_async_experiment_token = _current_experiment.set(self)
    self._event_loop = asyncio.get_event_loop()
    self._running_tasks = queue.Queue()
    self._work_unit_id_predictor = id_predictor.Predictor(
        1 + self.work_unit_count
    )
    return self

  async def _await_for_tasks(self):
    while not self._running_tasks.empty():
      await asyncio.wrap_future(self._running_tasks.get_nowait())

  async def __aexit__(self, exc_type, exc_value, traceback):  # pylint:disable=redefined-outer-name
    _current_experiment.reset(self._current_async_experiment_token)
    await self._await_for_tasks()

  @classmethod
  def package(
      cls, packageables: Sequence[job_blocks.Packageable] = ()
  ) -> Sequence[job_blocks.Executable]:
    """Packages `packageables` & triggers async packaging.

    This function has 2 usages:
    - Builds all given executables specs in parallel. While calling package(...)
      multiple times is allowed, that would result in slow sequential build,
      even if invoked from concurrent threads.
    - Triggers packaging of the items enqueued previously with `package_async`.

    Args:
      packageables: A sequence of extra packageables to build synchronously.

    Returns:
      A sequence of packaging results associated to `packageables` (same order).
    """
    return cls._async_packager.package(packageables)

  @classmethod
  def package_async(
      cls, packageable: job_blocks.Packageable
  ) -> Awaitable[job_blocks.Executable]:
    """Queues executable spec to be packaged into executable.

    If gathering all packageables for a single `package()` call is inconvenient,
    one may request packaging with `package_async` and later trigger the build
    for the whole batch with `package()`.

    Usage:

      if eval:
        eval_executable = experiment.package_async(xm.blaze_binary(...))
      if train:
        train_executable = experiment.package_async(xm.blaze_binary(...))

      experiment.package()  # Explicitly trigger packaging.

      jobs = {}
      if eval:
        jobs['eval'] = xm.job(await eval_executable, ...)
      if train:
        jobs['train'] = xm.job(await train_executable, ...)

    Args:
      packageable: Executable spec to package.

    Returns:
      An awaitable for the packaging result.
    """
    return cls._async_packager.add(packageable)

  @overload
  def add(
      self,
      job: AuxiliaryUnitJob,
      args: Optional[Mapping[str, Any]] = ...,
      *,  # parameters after “*” are keyword-only parameters
      identity: str = '',
  ) -> asyncio.Future[ExperimentUnit]:
    ...

  @overload
  def add(
      self,
      job: job_blocks.JobType,
      args: Optional[Mapping[str, Any]] = ...,
      *,  # parameters after “*” are keyword-only parameters
      role: WorkUnitRole = ...,
      identity: str = '',
  ) -> asyncio.Future[WorkUnit]:
    ...

  @overload
  def add(
      self,
      job: job_blocks.JobType,
      args: Optional[Mapping[str, Any]],
      *,  # parameters after “*” are keyword-only parameters
      role: ExperimentUnitRole,
      identity: str = '',
  ) -> asyncio.Future[ExperimentUnit]:
    ...

  @overload
  def add(
      self,
      job: job_blocks.JobType,
      args: Optional[Mapping[str, Any]] = ...,
      *,  # parameters after “*” are keyword-only parameters
      role: ExperimentUnitRole,
      identity: str = '',
  ) -> asyncio.Future[ExperimentUnit]:
    ...

  # The ExperimentUnit return type is determined by the role.
  def add(self, job, args=None, *, role=WorkUnitRole(), identity: str = ''):
    # pyformat: disable
    """Adds a Job / JobGroup to the experiment.

    A new Experiment Unit is created to run the job.

    Args:
      job: A Job or JobGroup to add.
      args: Keyword arguments to be passed to the job. For Job and JobGroup args
        are recursively expanded. For example,

        ```
        wu.add(
            JobGroup(agent=Job(...)),
            args={'agent': {'args': {'learning_rate': 0.1}}},
        )
        ```

        would update `args` field of a job `agent` in the group.
      role: The role of this unit in the experiment structure.
      identity: Optional unique experiment unit identifier within the
        experiment. If not empty, `add` adopts an 'add if not exists' behavior.
        If a unit with the given identity already exists it will be returned as
        is, without modifications. JobGenerators would still run to allow them
        to recover after preemption.

    Returns:
      An awaitable that would be fulfilled when the job is launched.
    """
    # pyformat: enable
    role = pattern_matching.match(
        pattern_matching.Case([AuxiliaryUnitJob], lambda job: job.role),
        pattern_matching.Case([Any], lambda job: role),
    )(job)
    experiment_unit_future = self._create_experiment_unit(args, role, identity)

    async def launch():
      experiment_unit = await experiment_unit_future
      try:
        await experiment_unit.add(job, args, identity=identity)
      except Exception:
        try:
          experiment_unit.stop(
              mark_as_failed=True,
              message=f'Work unit creation failed. {traceback.format_exc()}',
          )
        except Exception as stop_exception:  # pylint: disable=broad-except
          logging.error("Couldn't stop experiment unit: %s", stop_exception)
        raise
      return experiment_unit

    return asyncio.wrap_future(
        self._create_task(launch()), loop=self._event_loop
    )

  @abc.abstractmethod
  def _create_experiment_unit(
      self,
      args: Optional[Mapping[str, Any]],
      role: ExperimentUnitRole,
      identity: str,
  ) -> Awaitable[ExperimentUnit]:
    """Creates a new experiment unit.

    Synchronously starts the experiment unit creation, ensuring that IDs would
    be assigned in invocation order. The operation itself may run asynchronously
    in background.

    Args:
      args: Executable unit arguments, to be show as a part of hyper-parameter
        sweep.
      role: Executable unit role: whether to create a work or auxiliary unit.
      identity: Optional user-given experiment unit id.

    Returns:
      An awaitable to the creation result.
    """
    raise NotImplementedError

  def _create_task(self, task: Awaitable[Any]) -> futures.Future[Any]:
    if not self._event_loop.is_running():
      raise RuntimeError(
          'Event loop is not running. Have you entered Experiment context '
          'manager (e.g. with xm.create-experiment() as experiment:)?'
      )
    future = asyncio.run_coroutine_threadsafe(task, loop=self._event_loop)
    self._running_tasks.put_nowait(future)
    return future

  @property
  def work_unit_count(self) -> int:
    """Returns how many work units the experiment has."""
    raise NotImplementedError

  @abc.abstractmethod
  def work_units(self) -> Mapping[int, WorkUnit]:
    """Returns a mapping from work_unit_id to an instance of the work unit."""
    raise NotImplementedError

  @property
  def context(self) -> metadata_context.MetadataContext:
    """Returns metadata context for the experiment."""
    return metadata_context.MetadataContext(
        creator=getpass.getuser(),
        annotations=metadata_context.ContextAnnotations(),
    )


@abc.abstractmethod
def create_experiment(experiment_title: Optional[str] = None) -> Experiment:
  """Returns a concrete Experiment instance."""
  raise NotImplementedError


@abc.abstractmethod
def get_experiment(experiment_id: int) -> Experiment:
  """Returns an Experiment instance associated with this experiment id.

  Args:
    experiment_id: An ID of an experiment to get.

  Raises:
    NotFoundError: If experiment is not found.
  """
  raise NotImplementedError



import socket


def host_is_local(hostname, port=None):
    """returns True if the hostname points to the localhost, otherwise False."""
    if port is None:
        port = 22  # no port specified, lets just use the ssh port
    hostname = socket.getfqdn(hostname)
    if hostname in ("localhost", "0.0.0.0"):
        return True
    localhost = socket.gethostname()
    localaddrs = socket.getaddrinfo(localhost, port)
    targetaddrs = socket.getaddrinfo(hostname, port)
    for family, socktype, proto, canonname, sockaddr in localaddrs:
        for rfamily, rsocktype, rproto, rcanonname, rsockaddr in targetaddrs:
            if rsockaddr[0] == sockaddr[0]:
                return True
    return False


if __name__ == "__main__":
    i = input("Please enter an hostname or IP: ")
    il = host_is_local(i)
    if il:
        print("Hostname/IP is the localhost.")
    else:
        print("Hostname/IP is not the localhost.")



# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A module for communicating with the Bazel server."""

import abc
from typing import List, Sequence, Tuple

import attr


class BazelService(abc.ABC):
  """An interface for Bazel operations."""

  @abc.abstractmethod
  def fetch_kinds(self, labels: Sequence[str]) -> List[str]:
    """Fetches kinds of given targets.

    See https://docs.bazel.build/versions/main/query.html#output-label_kind.

    Args:
      labels: Labels of the targets to query.

    Returns:
      A list of kinds, for example, `['py_binary rule']`.
    """
    raise NotImplementedError

  @abc.abstractmethod
  def build_targets(
      self, labels: Sequence[str], tail_args: Sequence[str]
  ) -> List[List[str]]:
    """Builds given targets and returns paths to their important outputs.

    Args:
      labels: Labels of the targets to build.
      tail_args: Arguments to append to the Bazel command.

    Returns:
      For each label returns a list of its important outputs.
    """
    raise NotImplementedError


def _to_tuple(sequence: Sequence[str]) -> Tuple[str, ...]:
  """A standalone function to satisfy PyType."""
  return tuple(sequence)


@attr.s(auto_attribs=True, frozen=True)
class BazelTarget:
  """A Bazel target to be built."""

  label: str
  bazel_args: Tuple[str, ...] = attr.ib(converter=_to_tuple)



"""Utilities for automatically setting up UCL GridEngine executors.
This module includes utility for UCL users to automatically set up the
cluster-specific job requirements given more abstract requirements.

This module currently only works with the CS HPC cluster and myriad.
To use this, you need to set up the configuration file such that it includes
a cluster named "cs" and a cluster named "myriad".
An example configuration looks like the following:

```
project = ""
[local.storage]
staging = "/home/<your username>/.cache/lxm3"

[[clusters]]
name = "cs"
# Or any other login node.
server = "beaker.cs.ucl.ac.uk"
user = "<redacted>"

[clusters.storage]
staging = "/home/<your user name>/lxm3-staging"

[[clusters]]
name = "myriad"
server = "myriad.cs.ucl.ac.uk"
user = "<redacted>"

# staging must be a directory that is accessible from both the login node
# and compute node. For myriad, it is not correct to use
/home/<user>/Scratch as that is a symlink only visible from the login node.
[clusters.storage]
staging = "/scratch/scratch/<your user name>/lxm3-staging"

```

Then in your launcher script you can use something like the following:
```
requirements = xm_cluster.JobRequirements(gpu=1, ram=8 * xm.GB)
executor = ucl.UclGridEngine(requirements, walltime = 8 * xm.Hr)
```

This translates to the correct way to specify job resources '-l' for cs and myriad.
If you have configured both clusters, you can choose which cluster to use by
either setting requirements.location or using $LXM_CLUSTER environment variable.
If both are not set, the default is the first cluster in the configuration file.

To further customize the executor, you can manually modifer `executor.resources`
and `executor.parallel_environments` after the call to `ucl.UclGridEngine`.

To understand the rules for translating for the two clusters, refer to `:ucl_test`.

"""
from typing import Optional

from lxm3.xm_cluster import config
from lxm3.xm_cluster import executors as cluster_executors
from lxm3.xm_cluster import requirements as cluster_requirements


def _myriad_executor_fn(requirements: cluster_requirements.JobRequirements, **kwargs):
    resources = {}
    parallel_environments = {}
    for resource, value in requirements.task_requirements.items():
        if resource == cluster_requirements.ResourceType.RAM:
            resources["mem"] = value
        elif resource == cluster_requirements.ResourceType.GPU and value > 0:
            resources["gpu"] = value
        elif resource == cluster_requirements.ResourceType.CPU and value > 1:
            parallel_environments["smp"] = value

    return cluster_executors.GridEngine(
        resources=resources,
        parallel_environments=parallel_environments,
        modules=["singularity-env"],
        **kwargs,
    )


def _cs_executor_fn(requirements: cluster_requirements.JobRequirements, **kwargs):
    resources = {}
    parallel_environments = {}
    for resource, value in requirements.task_requirements.items():
        if resource == cluster_requirements.ResourceType.RAM:
            resources["h_vmem"] = value
            resources["tmem"] = value
        elif resource == cluster_requirements.ResourceType.GPU and value > 0:
            resources["gpu"] = "true"
            if value > 1:
                parallel_environments["gpu"] = value
        elif resource == cluster_requirements.ResourceType.CPU and value > 1:
            parallel_environments["smp"] = value

    if requirements.task_requirements.get(cluster_requirements.ResourceType.GPU, 0) > 0:
        resources.pop("h_vmem", None)

    return cluster_executors.GridEngine(
        resources=resources,
        parallel_environments=parallel_environments,
        **kwargs,
    )


def UclGridEngine(
    requirements: cluster_requirements.JobRequirements,
    walltime: Optional[int] = None,
    **kwargs,
) -> cluster_executors.GridEngine:
    """Create a UCL speicfic GridEngine executor.
    Args:
        requirements: The job requirements.
        walltime: The walltime in seconds.
        **kwargs: Additional arguments to pass to the executor.
        See xm_cluster.GridEngine.
    """
    executor_fns = {
        "myriad": _myriad_executor_fn,
        "cs": _cs_executor_fn,
    }
    location = requirements.location or config.default().default_cluster()
    if location not in executor_fns:
        raise ValueError(
            f"Unsupported location {location} for UCL GridEngine. Supported locations: "
            f"{list(executor_fns.keys())}"
        )
    executor = executor_fns[location](requirements, walltime=walltime, **kwargs)
    executor.requirements.location = location
    return executor



import datetime
import os
from typing import List, Optional, Union, cast

import paramiko

from lxm3 import xm
from lxm3.xm_cluster import config as config_lib
from lxm3.xm_cluster import executables
from lxm3.xm_cluster import executors
from lxm3.xm_cluster.execution import artifacts
from lxm3.xm_cluster.execution import job_script


def create_array_job(
    *,
    artifact: artifacts.Artifact,
    executable: executables.Command,
    singularity_image: Optional[str],
    singularity_options: Optional[executors.SingularityOptions],
    jobs: List[xm.Job],
    use_gpu: bool,
    version: Optional[str] = None,
    job_script_shebang: str,
    task_offset: int,
    task_id_var_name: str,
    setup: str,
    header: str,
):
    version = version or datetime.datetime.now().strftime("%Y%m%d.%H%M%S")

    job_name = f"job-{version}"
    array_wrapper = _create_array_wrapper(executable, jobs, task_offset)
    deploy_archive_path = artifact.archive_path(executable.resource_uri)
    setup_cmds = """
set -e

_prepare_workdir() {
    WORKDIR=$(mktemp -t -d -u lxm-workdir.XXXXX)
    echo >&2 "INFO[$(basename $0)]: Prepare work directory: $WORKDIR"
    mkdir -p $WORKDIR

    # Extract archives
    ARCHIVES="%(archives)s"
    for ar in $ARCHIVES; do
        unzip -q -d $WORKDIR $ar
    done

    ARRAY_WRAPPER_PATH=$(mktemp -t -u lxm-array-wrapper.XXXXX)
    echo >&2 "DEBUG[$(basename $0)]: Create wrapper script: $ARRAY_WRAPPER_PATH"
    cat << 'EOF' > $ARRAY_WRAPPER_PATH
%(array_wrapper)s
EOF
    chmod +x $ARRAY_WRAPPER_PATH

    _cleanup() {
        echo >&2 "INFO[$(basename $0)]: Clean up work directory: $WORKDIR"
        rm -rf $WORKDIR
        rm -f $ARRAY_WRAPPER_PATH
    }

    trap _cleanup EXIT

}

_prepare_workdir
cd $WORKDIR
%(setup)s
""" % {
        "array_wrapper": array_wrapper,
        "archives": " ".join([deploy_archive_path]),
        "setup": setup,
    }
    # Always use the array wrapper for but hardcode task id if not launching one job.
    if len(jobs) > 1:
        job_command = " ".join(["sh", "$ARRAY_WRAPPER_PATH", f"${task_id_var_name}"])
    else:
        job_command = " ".join(["sh", "$ARRAY_WRAPPER_PATH", f"{task_offset}"])

    if singularity_image is not None:
        deploy_container_path = artifact.singularity_image_path(
            os.path.basename(singularity_image)
        )
        singularity_options = singularity_options
        job_command = _wrap_singularity_cmd(
            job_command,
            deploy_container_path,
            singularity_options,
            use_gpu,
        )

    job_script_content = job_script.create_job_script(
        job_command,
        header,
        setup=setup_cmds,
        shebang=job_script_shebang,
    )

    return _put_job_resources(
        artifact=artifact,
        executable=executable,
        singularity_image=singularity_image,
        job_name=job_name,
        job_script_content=job_script_content,
    )


def _put_job_resources(
    *,
    artifact: artifacts.Artifact,
    executable: executables.Command,
    singularity_image: Optional[str],
    job_name: str,
    job_script_content: str,
) -> str:
    # Put artifacts on the staging fs
    job_script_dir = artifact.job_path(job_name)
    artifact.deploy_resource_archive(executable.resource_uri)

    if singularity_image is not None:
        artifact.deploy_singularity_container(singularity_image)

    artifact.deploy_job_scripts(job_name, job_script_content)

    deploy_job_script_path = os.path.join(job_script_dir, job_script.JOB_SCRIPT_NAME)

    return deploy_job_script_path


def _create_array_wrapper(
    executable: executables.Command, jobs: List[xm.Job], task_offset: int
):
    work_list = job_script.worklist_from_jobs(executable, jobs)
    return job_script.create_array_wrapper_script(
        cmd=executable.entrypoint_command, work_list=work_list, task_offset=task_offset
    )


def _wrap_singularity_cmd(
    job_command: str,
    deploy_container_path: str,
    singularity_options: Optional[executors.SingularityOptions],
    use_gpu: bool,
) -> str:
    singularity_opts = " ".join(
        job_script.get_singulation_options(singularity_options, use_gpu)
    )
    job_command = (
        f"singularity exec {singularity_opts} {deploy_container_path} {job_command}"
    )
    return job_command


def get_cluster_settings(config: config_lib.Config, jobs: List[xm.Job]):
    executor = cast(Union[executors.Slurm, executors.GridEngine], jobs[0].executor)
    location = executor.requirements.location

    for job in jobs:
        if not isinstance(job.executor, (executors.GridEngine, executors.Slurm)):
            raise ValueError("Only GridEngine and Slurm executors are supported.")
        if job.executor.requirements.location != location:
            raise ValueError("All jobs must be launched on the same cluster.")

    if location is None:
        location = config.default_cluster()

    cluster_config = config.cluster_config(location)
    storage_root = cluster_config["storage"]["staging"]
    hostname = cluster_config.get("server", None)
    user = cluster_config.get("user", None)

    connect_kwargs = {}
    proxycommand = cluster_config.get("proxycommand", None)
    if proxycommand is not None:
        connect_kwargs["sock"] = paramiko.ProxyCommand(proxycommand)

    ssh_private_key = cluster_config.get("ssh_private_key", None)
    if ssh_private_key is not None:
        connect_kwargs["key_filename"] = os.path.expanduser(ssh_private_key)

    password = cluster_config.get("password", None)
    if password is not None:
        connect_kwargs["password"] = password

    return storage_root, hostname, user, connect_kwargs


def write_job_id(artifact, job_script_path: str, job_id: str):
    artifact._fs.write_text(
        os.path.join(os.path.dirname(job_script_path), "job_id"), f"{job_id}\n"
    )



# noqa
import datetime
import os
from typing import List, Optional

from lxm3 import xm
from lxm3.clusters import slurm
from lxm3.xm_cluster import config as config_lib
from lxm3.xm_cluster import executables
from lxm3.xm_cluster import executors
from lxm3.xm_cluster.console import console
from lxm3.xm_cluster.execution import artifacts
from lxm3.xm_cluster.execution import common
from lxm3.xm_cluster.execution import job_script

_TASK_OFFSET = 1
_JOB_SCRIPT_SHEBANG = "#!/usr/bin/bash -l"
_TASK_ID_VAR_NAME = "SLURM_ARRAY_TASK_ID"


def _format_slurm_time(duration: datetime.timedelta) -> str:
    # See
    # https://github.com/SchedMD/slurm/blob/master/src/common/parse_time.c#L786
    days = duration.days
    seconds = int(duration.seconds)
    hours, remainder = divmod(seconds, 3600)
    minutes, seconds = divmod(remainder, 60)
    if days > 0:
        return "{:02}-{:02}:{:02}:{:02}".format(duration.days, hours, minutes, seconds)
    else:
        return "{:02}:{:02}:{:02}".format(hours, minutes, seconds)


def _generate_header_from_executor(
    job_name: str,
    executor: executors.Slurm,
    num_array_tasks: Optional[int],
    job_script_dir: str,
) -> str:
    header = []

    header.append(f"#SBATCH --job-name={job_name}")
    # TODO(yl): Only one task is supported for now.
    header.append("#SBATCH --ntasks=1")

    for resource, value in executor.resources.items():
        if value:
            header.append(f"#SBATCH --{resource}={value}")

    if executor.walltime is not None:
        duration = executor.walltime
        header.append(f"#SBATCH --time={_format_slurm_time(duration)}")

    log_directory = executor.log_directory or os.path.join(job_script_dir, "logs")
    if num_array_tasks is not None:
        stdout = os.path.join(log_directory, "slurm-%A_%a.out")
    else:
        stdout = os.path.join(log_directory, "slurm-%j.out")

    header.append(f"#SBATCH --output={stdout}")

    if executor.exclusive:
        header.append("#SBATCH --exclusive")

    if executor.partition:
        header.append(f"#SBATCH --partition={executor.partition}")

    if num_array_tasks is not None:
        array_spec = f"1-{num_array_tasks}"
        header.append(f"#SBATCH --array={array_spec}")

    # Skip requested header directives
    header = list(
        filter(
            lambda line: not any(skip in line for skip in executor.skip_directives),
            header,
        )
    )

    for line in executor.extra_directives:
        if not line.startswith("#SBATCH"):
            line = "#SBATCH " + line
        header.append(line)

    return "\n".join(header)


def _is_gpu_requested(executor: executors.Slurm) -> bool:
    return True  # TODO


def _create_job_header(
    executor: executors.Slurm, jobs: List[xm.Job], job_script_dir: str, job_name: str
) -> str:
    num_array_tasks = len(jobs) if len(jobs) > 1 else None
    job_header = _generate_header_from_executor(
        job_name, executor, num_array_tasks, job_script_dir
    )

    return job_header


def _get_setup_cmds(
    executable: executables.Command,
    executor: executors.Slurm,
) -> str:
    cmds = ["echo >&2 INFO[$(basename $0)]: Running on host $(hostname)"]

    for module in executor.modules:
        cmds.append(f"module load {module}")
    if _is_gpu_requested(executor):
        cmds.append(
            "echo >&2 INFO[$(basename $0)]: CUDA_VISIBLE_DEVICES=$CUDA_VISIBLE_DEVICES"
        )
        # cmds.append("nvidia-smi")

    if executable.singularity_image is not None:
        cmds.append(
            "echo >&2 INFO[$(basename $0)]: Singularity version: $(singularity --version)"
        )
    return "\n".join(cmds)


def deploy_job_resources(
    artifact: artifacts.Artifact,
    jobs: List[xm.Job],
    version: Optional[str] = None,
) -> str:
    version = version or datetime.datetime.now().strftime("%Y%m%d.%H%M%S")

    executable = jobs[0].executable
    executor = jobs[0].executor

    assert isinstance(executor, executors.Slurm)
    assert isinstance(executable, executables.Command)
    job_script.validate_same_job_configuration(jobs)

    job_name = f"job-{version}"

    job_script_dir = artifact.job_path(job_name)

    setup = _get_setup_cmds(executable, executor)
    header = _create_job_header(executor, jobs, job_script_dir, job_name)

    return common.create_array_job(
        artifact=artifact,
        executable=executable,
        singularity_image=executable.singularity_image,
        singularity_options=executor.singularity_options,
        jobs=jobs,
        use_gpu=_is_gpu_requested(executor),
        version=version,
        job_script_shebang=_JOB_SCRIPT_SHEBANG,
        task_offset=_TASK_OFFSET,
        task_id_var_name=_TASK_ID_VAR_NAME,
        setup=setup,
        header=header,
    )


class SlurmHandle:
    def __init__(self, job_id: str) -> None:
        self.job_id = job_id

    async def wait(self):
        raise NotImplementedError()

    async def monitor(self):
        raise NotImplementedError()


async def launch(config: config_lib.Config, jobs: List[xm.Job]) -> List[SlurmHandle]:
    if len(jobs) < 1:
        return []

    if not isinstance(jobs[0].executor, executors.Slurm):
        raise ValueError(
            "Only GridEngine executors are supported by the gridengine backend."
        )

    storage_root, hostname, user, connect_kwargs = common.get_cluster_settings(
        config, jobs
    )

    artifact = artifacts.create_artifact_store(
        storage_root,
        hostname=hostname,
        user=user,
        project=config.project(),
        connect_kwargs=connect_kwargs,
    )

    job_script_path = deploy_job_resources(artifact, jobs)

    console.log(f"Launch with command:\n  sbatch {job_script_path}")
    client = slurm.Client(hostname=hostname, username=user)

    job_id = client.launch(job_script_path)
    common.write_job_id(artifact, job_script_path, str(job_id))
    if len(jobs) > 1:
        job_ids = [f"{job_id}_{i}" for i in range(len(jobs))]
    else:
        job_ids = [f"{job_id}"]
    console.log(f"Successfully launched job {job_id}")

    return [SlurmHandle(j) for j in job_ids]



import fsspec
import paramiko

client = paramiko.SSHClient()
client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
fs = fsspec.filesystem(
    "sftp",
    host="slurm.ttic.edu",
    username="yicheng",
    key_filename="/home/yicheng/.ssh/id_ecdsa_ttic",
    sock=paramiko.ProxyCommand("ssh -W slurm.ttic.edu:22 yicheng@ssh.ttic.edu"),
)

print(fs)



#!/usr/bin/env python3
import argparse
import errno
import importlib
import os
import sys

from absl import app
from absl.flags import argparse_flags

import lxm3


def version(_):
    print(f"lxm3 {lxm3.__version__}")


def register_version_parser(parsers: argparse._SubParsersAction):
    version_parser = parsers.add_parser(
        "version",
        help="Print version.",
        inherited_absl_flags=None,  # type: ignore
    )
    version_parser.set_defaults(command=version)


def launch(args):
    launch_script = args.launch_script
    if not os.path.exists(launch_script):
        raise OSError(errno.ENOENT, f"File not found: {launch_script}")
    sys.path.insert(0, os.path.abspath(os.path.dirname(launch_script)))
    launch_module, _ = os.path.splitext(os.path.basename(launch_script))
    m = importlib.import_module(launch_module)
    argv = [launch_script, "--xm_launch_script={}".format(launch_script)] + args.args
    app.run(m.main, argv=argv)
    sys.path.pop(0)


def shell(args):
    del args
    from lxm3.cli.shell import main

    app.run(main)


def register_launch_parser(parsers: argparse._SubParsersAction):
    launch_parser = parsers.add_parser(
        "launch",
        help="Launch experiment.",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        inherited_absl_flags=None,  # type: ignore
        epilog=r"""
examples:

  Launch experiment defined in "launcher.py"
  and pass extra args "--task 1" to "launcher.py".

    lxm3 launch launcher.py -- --task 1
""",
    )
    launch_parser.add_argument(
        "launch_script",
        metavar="LAUNCH_SCRIPT",
        # nargs=1,
        help="Path to launch script.",
    )
    launch_parser.add_argument(
        "args",
        nargs=argparse.REMAINDER,
        metavar="ARGS",
        help="Additional arguments to pass to launch script.",
    )
    launch_parser.set_defaults(command=launch)


def register_shell_parser(parsers: argparse._SubParsersAction):
    launch_parser = parsers.add_parser(
        "shell",
        help="Open a shell.",
        inherited_absl_flags=None,  # type: ignore
    )
    launch_parser.set_defaults(command=shell)


def _parse_flags(argv):
    parser = argparse_flags.ArgumentParser(description="lxm3 experiment scheduler.")
    parser.set_defaults(command=lambda _: parser.print_help())

    subparsers = parser.add_subparsers()

    register_version_parser(subparsers)
    register_launch_parser(subparsers)
    register_shell_parser(subparsers)

    args = parser.parse_args(argv[1:])
    return args


def main(args):
    args.command(args)


def entrypoint():
    app.run(main, flags_parser=_parse_flags)


if __name__ == "__main__":
    entrypoint()



# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Resources specification for use in the API.

Various classes defined to support resources specification for jobs.
"""

import enum
import functools
import itertools
import operator
import re
from typing import Any, Iterable, Iterator, Dict, Mapping, MutableMapping, Optional, Tuple, Union

import immutabledict
from lxm3._vendor.xmanager.xm import pattern_matching as pm


class _CaseInsensetiveResourceTypeMeta(enum.EnumMeta):
  """Metaclass which allows case-insensetive enum lookup.

  Enum keys are upper case, but we allow other cases for the input. For
  example existing flags and JobRequirements use lower case for resource names.
  """

  def __getitem__(cls, resource_name: str) -> 'ResourceType':
    try:
      return super().__getitem__(resource_name.upper())
    except KeyError:
      raise KeyError(f'Unknown {cls.__name__} {resource_name!r}')  # pylint: disable=raise-missing-from


class ResourceType(enum.Enum, metaclass=_CaseInsensetiveResourceTypeMeta):
  """Type of a countable resource (e.g., CPU, memory, accelerators etc).

  We use a schema in which every particular accelerator has its own type. This
  way all countable resources required for a job could be represented by a
  simple dictionary.
  """

  # Amount of required CPU resources in vCPUs.
  CPU = 100002
  # Amount of required memory resources in bytes.
  MEMORY = 39
  RAM = 39
  # Amount of required disk resources in bytes.
  EPHEMERAL_STORAGE = 100003
  DISK = 100003

  # GPUs
  LOCAL_GPU = 100006
  P4 = 21
  T4 = 22
  P100 = 14
  V100 = 17
  A100 = 46
  A100_80GIB = 66

  # TPUs
  TPU_V2 = 3
  TPU_V3 = 16

  # TODO: do we need V2_DONUT and V3_DONUT?

  def __str__(self):
    return self._name_


class _CaseInsensetiveServiceTierMeta(enum.EnumMeta):
  """Metaclass which allows case-insensetive enum lookup.

  Enum keys are upper case, but we allow other cases for the input. For
  example existing flags and JobRequirements use lower case for resource names.
  """

  def __getitem__(cls, resource_name: str) -> 'ServiceTier':
    try:
      return super().__getitem__(resource_name.upper())
    except KeyError:
      raise KeyError(f'Unknown {cls.__name__} {resource_name!r}')  # pylint: disable=raise-missing-from


class ServiceTier(enum.Enum, metaclass=_CaseInsensetiveServiceTierMeta):
  """The job availability guarantees which underlying platfrom should provide.

  Most cloud platforms offer a selection of availability/price tradeoff options.
  Usually there are at least "Take my money, this workload is important" and
  "Buy excess compute for cheap" offerings. This enum provides a classification
  of such offerings and allows matching comparable (but not necessary identical)
  options from different runtimes.
  """

  # Highly available tier. The job is expected to be scheduled fast once sent to
  # the cloud. Recommended tier for multi-job work units as lower tiers may lead
  # to partially-scheduled work units.
  PROD = 200
  # A cheaper tier with guaranteed average throughput. Jobs may spend hours
  # awaiting scheduling by the cloud and can be preempted.
  BATCH = 100
  # The cheapest tier. No guarantees, but it often works.
  FREEBIE = 25


def _enum_subset(class_name: str, values: Iterable[ResourceType]) -> type:  # pylint: disable=g-bare-generic
  """Returns an enum subset class.

  The class is syntactically equivalent to an enum with the given resource
  types. But the concrete constants are the same as in the ResourceType enum,
  making all equivalence comparisons work correctly. Additionally operator `in`
  is supported for checking if a resource belongs to the subset.

  Args:
    class_name: Class name of the subset enum.
    values: A list of resources that belong to the subset.
  """
  values = set(values)

  class EnumSubsetMetaclass(type):  # pylint: disable=g-bare-generic
    """Metaclass which implements enum subset operations."""

    def __new__(
        cls,
        name: str,
        bases: Tuple[type],  # pylint: disable=g-bare-generic
        dct: Dict[str, Any],
    ) -> type:  # pylint: disable=g-bare-generic
      # Add constants to the class dict.
      for name, member in ResourceType.__members__.items():
        if member in values:
          dct[name] = member

      return super().__new__(cls, class_name, bases, dct)

    def __getitem__(cls, item: str) -> ResourceType:
      result = ResourceType[item]
      if result not in cls:  # pylint: disable=unsupported-membership-test
        raise AttributeError(
            f"type object '{cls.__name__}' has no attribute '{item}'"
        )
      return result

    def __iter__(cls) -> Iterator[ResourceType]:
      return iter(values)

    def contains(cls, value: ResourceType) -> bool:
      return value in values

  class EnumSubset(metaclass=EnumSubsetMetaclass):

    def __new__(cls, value: int) -> ResourceType:
      resource = ResourceType(value)
      if resource not in cls:
        raise ValueError(f'{value} is not a valid {cls.__name__}')
      return resource

  return EnumSubset


# TODO: Use centralized resource metadata.
TpuType = _enum_subset(
    'TpuType',
    [
        ResourceType.TPU_V2,
        ResourceType.TPU_V3,
    ],
)


GpuType = _enum_subset(
    'GpuType',
    [
        # LOCAL_GPU is missing as only specific GPU types should be added.
        ResourceType.P4,
        ResourceType.T4,
        ResourceType.P100,
        ResourceType.V100,
        ResourceType.A100,
        ResourceType.A100_80GIB,
    ],
)


_AcceleratorType = _enum_subset(
    '_AcceleratorType',
    [
        ResourceType.LOCAL_GPU,
        *list(TpuType),
        *list(GpuType),
    ],
)


class ResourceDict(MutableMapping):
  """Internal class to represent amount of countable resources.

  A mapping from ResourceType to amount of the resource combined with
  convenience methods. This class only tracks amounts of the resources, but not
  their topologies, locations or constraints.

  This class is rather generic and is designed be used internally as job
  requirements as well as in the executors. API users should not use it
  explicitly.

  Usage:
    # Construct (implicitly) from user code using JobRequirements:
    requirements = JobRequirements(cpu=0.5 * xm.vCPU, memory=2 * xm.GiB, v100=8)
    resources = requirements.task_requirements
    # Resources are available by their canonical names.
    assert(resources[ResourceType.V100], 8)
    # Print user-friendly representation:
    print(f'The task needs {resources}')
  """

  def __init__(self) -> None:
    self.__dict: Dict[ResourceType, float] = {}

  def __setitem__(self, key: ResourceType, value: float) -> None:
    self.__dict.__setitem__(key, value)

  def __getitem__(self, key: ResourceType) -> float:
    return self.__dict.__getitem__(key)

  def __delitem__(self, key: ResourceType) -> None:
    self.__dict.__delitem__(key)

  def __iter__(self):
    return self.__dict.__iter__()

  def __len__(self) -> int:
    return self.__dict.__len__()

  def __str__(self) -> str:
    """Returns user-readable text representation.

    Such as "V100: 8, CPU: 1.2, MEMORY: 5.4GiB".
    """
    # TODO: We do not aggregate memory yet, update this method to be more
    # user-friendly.
    return ', '.join(
        sorted([f'{key}: {value}' for (key, value) in self.items()])
    )

  def __add__(self: 'ResourceDict', rhs: 'ResourceDict') -> 'ResourceDict':
    """Returns a sum of two ResourceDicts."""
    result = ResourceDict()
    for key in [*self.keys(), *rhs.keys()]:
      result[key] = self.get(key, 0) + rhs.get(key, 0)
    return result

  def __mul__(self: 'ResourceDict', rhs: float) -> 'ResourceDict':
    """Returns the multiplication of a ResourceDict with a scalar."""
    result = ResourceDict()
    for key, value in self.items():
      result[key] = value * rhs
    return result

  def __rmul__(self: 'ResourceDict', rhs: float) -> 'ResourceDict':
    """Returns the multiplication of a ResourceDict with a scalar."""
    return self * rhs


class InvalidTpuTopologyError(Exception):
  """An unrecognized TPU topology has been provided."""


class Topology:
  """Accelerator topology configuration.

  Describes accelerator interconnection. For example could be a TPU topology or
  several GPUs connected with NVLink. Topologies have a form of 'NxM_suffix'
  where N & M are the number of accelerators across the dimension and suffix
  corresponds to a specific interconnect type. Number of dimensions may vary.

  Examples of valid topologies:
    '1' - a single device.
    '4' - 4 GPUs on one host.
    '4x4' - a 4x4 TPU grid.
  """

  def __init__(self, name: str) -> None:
    if not re.fullmatch('([\\d]+x?)+(_(un)?twisted)?', name):
      raise InvalidTpuTopologyError(f'Invalid TPU topology: {name}.')

    self._name = name

    dimensions_str = name.split('_')[0]
    self.dimensions = list(map(int, dimensions_str.split('x')))

  @property
  def chip_count(self) -> int:
    """Returns the number of chips of the TPU topology."""
    return functools.reduce(operator.mul, self.dimensions)

  @property
  def name(self) -> str:
    """Returns the topology as a string."""
    return self._name

  def __repr__(self) -> str:
    return f'xm.Topology({self.name!r})'

  def __eq__(self, other: 'Topology') -> bool:
    return self.name == other.name

  def __hash__(self) -> int:
    return hash(self.name)


ResourceQuantity = Union[int, float, str, Topology]


def _parse_resource_quantity(
    resource_name: str, value: ResourceQuantity
) -> Tuple[float, Optional[Topology]]:
  """Parses a string representation of a resource quantity."""

  def parse_string(value: str):
    if 'x' in value:
      topology = Topology(value)
      return topology.chip_count, topology
    else:
      # TODO: Parse SI suffixes, like GiB.
      return float(value), None

  def parse_topology(topology: Topology):
    return topology.chip_count, topology

  def parse_number(value: Any):
    return float(value), None

  try:
    return pm.match(parse_string, parse_topology, parse_number)(value)
  except Exception as e:
    raise ValueError(
        f"Couldn't parse resource quantity for {resource_name}. "
        f'{value!r} was given.'
    ) from e


class JobRequirements:
  # pyformat: disable
  """Describes the resource requirements of a Job.

  Attributes:
    task_requirements: Amount of resources needed for a single task within a
      job.
    accelerator: The accelerator the jobs uses, if there is one. Jobs using
      multiple accelerators are not supported because different kinds of
      accelerators are usually not installed on the same host.
    topology: Accelerator topology, if an accelerator is used.
    location: Place where the job should run. For example a cluster name or a
      Borg cell.
    service_tier: A service tier at which the job should run.
    replicas: Number of identical tasks to run within a job
  """
  # pyformat:enable

  task_requirements: ResourceDict
  accelerator: Optional[ResourceType]
  topology: Optional[Topology]

  location: Optional[str]
  _service_tier: ServiceTier

  def __init__(
      self,
      resources: Mapping[
          Union[ResourceType, str], ResourceQuantity
      ] = immutabledict.immutabledict(),
      *,
      location: Optional[str] = None,
      replicas: Optional[int] = None,
      service_tier: Optional[ServiceTier] = None,
      **kw_resources: ResourceQuantity,
  ) -> None:
    # pyformat: disable
    """Define a set of resources.

    Args:
      resources: resource amounts as a dictionary, for example
        {xm.ResourceType.V100: 2}.
      location: Place where the job should run. For example a cluster name or a
        Borg cell.
      replicas: Number of identical tasks to run within a job. 1 by default.
      service_tier: A service tier at which the job should run.
      **kw_resources: resource amounts as a kwargs, for example `v100=2` or
        `ram=1 * xm.GiB`. See xm.ResourceType enum for the list of supported
        types and aliases.

    Raises:
      ValueError:
        If several accelerator resources are supplied (i.e. GPU and TPU).
        If the same resource is passed in a `resources` dictionary and as
          a command line argument.
        If topology is supplied for a non accelerator resource.
    """
    # pyformat: enable
    self.location = location
    self._service_tier = service_tier or ServiceTier.PROD

    self.task_requirements = ResourceDict()
    self.accelerator = None
    self.topology = None

    for resource_name, value in itertools.chain(
        resources.items(), kw_resources.items()
    ):
      scalar, topology = _parse_resource_quantity(resource_name, value)
      resource = pm.match(
          pm.Case([str], lambda r: ResourceType[r]),
          pm.Case([ResourceType], lambda r: r),
      )(resource_name)

      if resource in _AcceleratorType:
        if self.accelerator is not None:
          raise ValueError('Accelerator already set.')
        self.accelerator = resource
        self.topology = topology or Topology(f'{scalar:g}')
      elif topology is not None:
        raise ValueError(
            f'A topology specified for non accelerator resource {resource}.'
        )

      if resource in self.task_requirements:
        raise ValueError(f'{resource} has been specified twice.')
      self.task_requirements[resource] = scalar

    if (
        self.accelerator in GpuType
        and self.topology
        and len(self.topology.dimensions) == 2
    ):
      if replicas is not None and replicas != self.topology.dimensions[1]:
        raise ValueError(
            f'For multihost GPUs with topology {self.topology}, replicas should'
            f'be either None or {self.topology.dimensions[1]}. Found: '
            f'{replicas}'
        )
      replicas = self.topology.dimensions[1]

    self.replicas = replicas or 1
    self._validate_replicas()

  @property
  def service_tier(self):
    return self._service_tier

  @service_tier.setter
  def service_tier(self, new_service_tier):
    self._service_tier = new_service_tier

  def _validate_replicas(self) -> None:
    """Raises ValueError if replication is not supported."""
    if self.replicas > 1 and self.accelerator in TpuType:
      raise ValueError(
          f'Replicated jobs are not supported for {self.accelerator}.'
      )

  def __repr__(self) -> str:
    """Returns string representation of the requirements."""
    args = []

    for resource, value in self.task_requirements.items():
      if resource in TpuType:
        args.append(f'{resource.name.lower()}={self.topology!r}')
      else:
        args.append(f'{resource.name.lower()}={value!r}')

    if self.location:
      args.append(f'location={self.location!r}')
    if self.service_tier != ServiceTier.PROD:
      args.append(f'service_tier=xm.{self.service_tier}')
    if self.replicas != 1:
      args.append(f'replicas={self.replicas}')

    return f'xm.JobRequirements({", ".join(args)})'



import atexit
import datetime
import functools
import glob
import os
import shutil
import subprocess
import tempfile
from typing import Any, Sequence

from absl import logging

from lxm3 import xm
from lxm3._vendor.xmanager.xm import pattern_matching
from lxm3.xm_cluster import executable_specs as cluster_executable_specs
from lxm3.xm_cluster import executables as cluster_executables
from lxm3.xm_cluster.console import console

_ENTRYPOINT = "./entrypoint.sh"


class PackagingError(Exception):
    """Error raised when packaging fails."""


@functools.lru_cache()
def _staging_directory():
    staging_dir = tempfile.mkdtemp(prefix="xm_cluster_staging_")
    logging.debug("Created local staging directory: %s", staging_dir)

    def remove_staging_dir():
        logging.debug("Removing local staging directory: %s", staging_dir)
        shutil.rmtree(staging_dir)

    atexit.register(remove_staging_dir)
    return staging_dir


def _create_archive(
    staging_directory: str, py_package: cluster_executable_specs.PythonPackage
):
    package_name = py_package.name
    version = datetime.datetime.now().strftime("%Y%m%d.%H%M%S")
    archive_name = f"{package_name}-{version}"
    package_dir = py_package.path
    resources = py_package.resources

    with tempfile.TemporaryDirectory() as tmpdir:
        with console.status(
            "Creating python package archive for {}".format(package_dir)
        ):
            try:
                subprocess.run(
                    [
                        "pip",
                        "install",
                        *py_package.pip_args,
                        "-t",
                        tmpdir,
                        package_dir,
                        *py_package.extra_packages,
                    ],
                    text=True,
                    check=True,
                    capture_output=True,
                )
            except subprocess.CalledProcessError as e:
                raise PackagingError(
                    f"Failed to create python package from {package_dir}"
                ) from e

            # Add resources to the archive
            for resource in resources:
                for src, dst in resource.files:
                    target_file = os.path.join(tmpdir, dst)
                    if not os.path.exists(os.path.dirname(target_file)):
                        os.makedirs(os.path.dirname(target_file))
                    if not os.path.exists(target_file):
                        shutil.copy(src, target_file)
                    else:
                        raise ValueError(
                            "Additional resource overwrites existing file: %s", src
                        )

            if os.path.exists(os.path.join(tmpdir, "bin")):
                console.log('Removing "bin/" directory as these are not yet portable.')
                shutil.rmtree(os.path.join(tmpdir, "bin"))

            for f in glob.glob(os.path.join(tmpdir, "*.dist-info")):
                shutil.rmtree(f)

            entrypoint = f"""\
#!/bin/bash
export PYTHONPATH=$(dirname $0):$PYTHONPATH
{_create_entrypoint_cmds(py_package)}
"""
            with open(os.path.join(tmpdir, _ENTRYPOINT), "w") as f:
                f.write(entrypoint)
            os.chmod(f.name, 0o755)
            archive_name = shutil.make_archive(
                os.path.join(staging_directory, archive_name),
                "zip",
                tmpdir,
                verbose=True,
            )

        console.log(
            f"Created archive: [repr.path]{os.path.basename(archive_name)}[repr.path]"
        )
        return os.path.basename(archive_name)


def _create_entrypoint_cmds(python_package: cluster_executable_specs.PythonPackage):
    if isinstance(python_package.entrypoint, cluster_executable_specs.ModuleName):
        cmds = ["python3 -m {}".format(python_package.entrypoint.module_name)]
    elif isinstance(python_package.entrypoint, cluster_executable_specs.CommandList):
        cmds = python_package.entrypoint.commands
    else:
        raise ValueError("Unexpected entrypoint: {}".format(python_package.entrypoint))
    cmds = "\n".join(cmds)
    # Allow passing extra parameters to the commands.
    if not cmds.endswith(("$@", '"$@"')):
        cmds = cmds + ' "$@"'
    return cmds


def _package_python_package(
    py_package: cluster_executable_specs.PythonPackage,
    packageable: xm.Packageable,
):
    staging = tempfile.mkdtemp(dir=_staging_directory())
    archive_name = _create_archive(staging, py_package)
    local_archive_path = os.path.join(staging, archive_name)
    entrypoint_cmd = _ENTRYPOINT

    return cluster_executables.Command(
        entrypoint_command=entrypoint_cmd,
        resource_uri=local_archive_path,
        name=py_package.name,
        args=packageable.args,
        env_vars=packageable.env_vars,
    )


def _package_universal_package(
    universal_package: cluster_executable_specs.UniversalPackage,
    packageable: xm.Packageable,
):
    staging = tempfile.mkdtemp(dir=_staging_directory())
    version = datetime.datetime.now().strftime("%Y%m%d.%H%M%S")
    archive_name = f"{universal_package.name}-{version}"

    with tempfile.TemporaryDirectory(
        prefix=f"build-{archive_name}",
        dir=_staging_directory(),
    ) as build_dir:
        package_dir = universal_package.path
        build_env = {**os.environ, "BUILDDIR": build_dir}

        build_script = os.path.join(package_dir, universal_package.build_script)

        try:
            subprocess.run(
                [build_script] + universal_package.build_args,
                cwd=package_dir,
                env=build_env,
                check=True,
            )
        except subprocess.CalledProcessError as e:
            raise PackagingError(
                "Failed to build universal package from {}".format(package_dir)
            ) from e

        archive_name = shutil.make_archive(
            os.path.join(staging, archive_name), "zip", build_dir, verbose=True
        )

    return cluster_executables.Command(
        # TODO(yl): this is not very robust
        entrypoint_command=" ".join(universal_package.entrypoint),
        resource_uri=os.path.join(staging, os.path.basename(archive_name)),
        name=universal_package.name,
        args=packageable.args,
        env_vars=packageable.env_vars,
    )


def _package_singularity_container(
    container: cluster_executable_specs.SingularityContainer,
    packageable: xm.Packageable,
):
    executable = _PACKAGING_ROUTER(container.entrypoint, packageable)
    executable.singularity_image = container.image_path
    return executable


def _throw_on_unknown_executable(
    executable: Any,
    packageable: xm.Packageable,
):
    raise TypeError(
        f"Unsupported executable specification: {executable!r}. "
        f"Packageable: {packageable!r}"
    )


_PACKAGING_ROUTER = pattern_matching.match(
    _package_python_package,
    _package_universal_package,
    _package_singularity_container,
    _throw_on_unknown_executable,
)


def package(packageables: Sequence[xm.Packageable]):
    return [_PACKAGING_ROUTER(pkg.executable_spec, pkg) for pkg in packageables]



import getpass
from typing import Set

from lxm3._vendor.xmanager.xm import metadata_context


class ClusterContextAnnotations(metadata_context.ContextAnnotations):
    def __init__(self) -> None:
        self._title = ""
        self._tags = set()
        self._notes = ""

    @property
    def title(self) -> str:
        return self._title

    def set_title(self, title: str) -> None:
        self._title = title

    @property
    def tags(self) -> Set[str]:
        return self._tags

    def add_tags(self, *tags: str) -> None:
        self._tags.update(tags)

    def remove_tags(self, *tags: str) -> None:
        for tag in tags:
            self._tags.discard(tag)

    @property
    def notes(self) -> str:
        return self._notes

    def set_notes(self, notes: str) -> None:
        self._notes = notes


class ClusterMetadataContext(metadata_context.MetadataContext):
    def __init__(self) -> None:
        super().__init__(
            creator=getpass.getuser(), annotations=ClusterContextAnnotations()
        )



# type: ignore
import enum
import itertools
from typing import (
    Any,
    Dict,
    Iterable,
    Iterator,
    Mapping,
    MutableMapping,
    Optional,
    Tuple,
    Union,
)

import immutabledict

from lxm3._vendor.xmanager.xm import pattern_matching as pm


class _CaseInsensetiveResourceTypeMeta(enum.EnumMeta):
    """Metaclass which allows case-insensetive enum lookup.

    Enum keys are upper case, but we allow other cases for the input. For
    example existing flags and JobRequirements use lower case for resource names.
    """

    def __getitem__(cls, resource_name: str) -> "ResourceType":
        try:
            return super().__getitem__(resource_name.upper())
        except KeyError:
            raise KeyError(
                f"Unknown {cls.__name__} {resource_name!r}"
            )  # pylint: disable=raise-missing-from


class ResourceType(enum.Enum, metaclass=_CaseInsensetiveResourceTypeMeta):
    """Type of a countable resource (e.g., CPU, memory, accelerators etc).

    We use a schema in which every particular accelerator has its own type. This
    way all countable resources required for a job could be represented by a
    simple dictionary.
    """

    CPU = 100002
    RAM = 39

    # GPUs
    GPU = 100006

    # TODO: do we need V2_DONUT and V3_DONUT?

    def __str__(self):
        return self._name_


def _enum_subset(
    class_name: str, values: Iterable[ResourceType]
) -> type:  # pylint: disable=g-bare-generic
    """Returns an enum subset class.

    The class is syntactically equivalent to an enum with the given resource
    types. But the concrete constants are the same as in the ResourceType enum,
    making all equivalence comparisons work correctly. Additionally operator `in`
    is supported for checking if a resource belongs to the subset.

    Args:
      class_name: Class name of the subset enum.
      values: A list of resources that belong to the subset.
    """
    values = set(values)

    class EnumSubsetMetaclass(type):  # pylint: disable=g-bare-generic
        """Metaclass which implements enum subset operations."""

        def __new__(
            cls,
            name: str,
            bases: Tuple[type],  # pylint: disable=g-bare-generic
            dct: Dict[str, Any],
        ) -> type:  # pylint: disable=g-bare-generic
            # Add constants to the class dict.
            for name, member in ResourceType.__members__.items():
                if member in values:
                    dct[name] = member

            return super().__new__(cls, class_name, bases, dct)

        def __getitem__(cls, item: str) -> ResourceType:
            result = ResourceType[item]
            if result not in cls:  # pylint: disable=unsupported-membership-test
                raise AttributeError(
                    f"type object '{cls.__name__}' has no attribute '{item}'"
                )
            return result

        def __iter__(cls) -> Iterator[ResourceType]:
            return iter(values)

        def contains(cls, value: ResourceType) -> bool:
            return value in values

    class EnumSubset(metaclass=EnumSubsetMetaclass):
        def __new__(cls, value: int) -> ResourceType:
            resource = ResourceType(value)
            if resource not in cls:
                raise ValueError(f"{value} is not a valid {cls.__name__}")
            return resource

    return EnumSubset


_AcceleratorType = _enum_subset(
    "_AcceleratorType",
    [
        ResourceType.GPU,
    ],
)


class ResourceDict(MutableMapping):
    """Internal class to represent amount of countable resources.

    A mapping from ResourceType to amount of the resource combined with
    convenience methods. This class only tracks amounts of the resources, but not
    their topologies, locations or constraints.

    This class is rather generic and is designed be used internally as job
    requirements as well as in the executors. API users should not use it
    explicitly.

    Usage:
      # Construct (implicitly) from user code using JobRequirements:
      requirements = JobRequirements(cpu=0.5 * xm.vCPU, memory=2 * xm.GiB, v100=8)
      resources = requirements.task_requirements
      # Resources are available by their canonical names.
      assert(resources[ResourceType.V100], 8)
      # Print user-friendly representation:
      print(f'The task needs {resources}')
    """

    def __init__(self) -> None:
        self.__dict: Dict[ResourceType, float] = {}

    def __setitem__(self, key: ResourceType, value: float) -> None:
        self.__dict.__setitem__(key, value)

    def __getitem__(self, key: ResourceType) -> float:
        return self.__dict.__getitem__(key)

    def __delitem__(self, key: ResourceType) -> None:
        self.__dict.__delitem__(key)

    def __iter__(self):
        return self.__dict.__iter__()

    def __len__(self) -> int:
        return self.__dict.__len__()

    def __str__(self) -> str:
        """Returns user-readable text representation.

        Such as "V100: 8, CPU: 1.2, MEMORY: 5.4GiB".
        """
        # TODO: We do not aggregate memory yet, update this method to be more
        # user-friendly.
        return ", ".join(sorted([f"{key}: {value}" for (key, value) in self.items()]))

    def __add__(self: "ResourceDict", rhs: "ResourceDict") -> "ResourceDict":
        """Returns a sum of two ResourceDicts."""
        result = ResourceDict()
        for key in [*self.keys(), *rhs.keys()]:
            result[key] = self.get(key, 0) + rhs.get(key, 0)
        return result

    def __mul__(self: "ResourceDict", rhs: float) -> "ResourceDict":
        """Returns the multiplication of a ResourceDict with a scalar."""
        result = ResourceDict()
        for key, value in self.items():
            result[key] = value * rhs
        return result

    def __rmul__(self: "ResourceDict", rhs: float) -> "ResourceDict":
        """Returns the multiplication of a ResourceDict with a scalar."""
        return self * rhs


ResourceQuantity = Union[int, str]


class JobRequirements:
    # pyformat: disable
    """Describes the resource requirements of a Job.

    Attributes:
      task_requirements: Amount of resources needed for a single task within a
        job.
      accelerator: The accelerator the jobs uses, if there is one. Jobs using
        multiple accelerators are not supported because different kinds of
        accelerators are usually not installed on the same host.
      topology: Accelerator topology, if an accelerator is used.
      location: Place where the job should run. For example a cluster name or a
        Borg cell.
      service_tier: A service tier at which the job should run.
      replicas: Number of identical tasks to run within a job
    """
    # pyformat:enable

    task_requirements: ResourceDict
    accelerator: Optional[ResourceType]

    location: Optional[str]

    def __init__(
        self,
        resources: Mapping[
            Union[ResourceType, str], ResourceQuantity
        ] = immutabledict.immutabledict(),
        *,
        location: Optional[str] = None,
        **kw_resources: ResourceQuantity,
    ) -> None:
        # pyformat: disable
        """Define a set of resources.

        Args:
          resources: resource amounts as a dictionary, for example
            {xm.ResourceType.V100: 2}.
          location: Place where the job should run. For example a cluster name or a
            Borg cell.
          **kw_resources: resource amounts as a kwargs, for example `v100=2` or
            `ram=1 * xm.GiB`. See xm.ResourceType enum for the list of supported
            types and aliases.

        Raises:
          ValueError:
            If several accelerator resources are supplied (i.e. GPU and TPU).
            If the same resource is passed in a `resources` dictionary and as
              a command line argument.
            If topology is supplied for a non accelerator resource.
        """
        # pyformat: enable
        self.location = location

        self.task_requirements = ResourceDict()
        self.accelerator = None
        self.topology = None

        for resource_name, value in itertools.chain(
            resources.items(), kw_resources.items()
        ):
            resource = pm.match(
                pm.Case([str], lambda r: ResourceType[r]),
                pm.Case([ResourceType], lambda r: r),
            )(resource_name)

            if resource in _AcceleratorType:
                if self.accelerator is not None:
                    raise ValueError("Accelerator already set.")
                self.accelerator = resource

            if resource in self.task_requirements:
                raise ValueError(f"{resource} has been specified twice.")
            self.task_requirements[resource] = value

    def __repr__(self) -> str:
        """Returns string representation of the requirements."""
        args = []

        if self.location:
            args.append(f"location={self.location!r}")

        return f'xm.JobRequirements({", ".join(args)})'



# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Pattern matching implementation.

Until Python 3.10 with PEP 646 is available to us, this module provides an
alternative implementation of https://en.wikipedia.org/wiki/Pattern_matching.

This module is private and can only be used by the API itself, as the provided
API is not final.
"""

import inspect
from typing import Any, Awaitable, Callable, Iterable, Generic, Tuple, Type, TypeVar, Union

R = TypeVar('R')


class Case(Generic[R]):
  """A handler of a particular type branch."""

  # A tuple of types to match. One for each callable argument.
  # Immutable, should not be modified.
  kind: Tuple[Type[Any]]
  # A handler to be called in case of a match.
  # Must be `Callable[..., R]`. Immutable, should not be modified.
  handle: Any

  def __init__(
      self,
      kind: Iterable[Type[Any]],
      handler: Callable[..., R],
  ) -> None:
    """Constructs a handler for given types.

    Args:
      kind: A tuple of types to match. One for each callable argument.
      handler: A handler to be called in case of a match.
    """
    self.kind = tuple(
        getattr(arg_type, '__origin__', arg_type) for arg_type in kind
    )
    self.handle = handler

  def matches(self, *values: Any) -> bool:
    """Checks if *values match the types."""
    if len(values) != len(self.kind):
      return False

    return all(
        expected_type is Any or isinstance(value, expected_type)
        for expected_type, value in zip(self.kind, values)
    )


def _deduce_types(handler: Callable[..., Any]) -> Tuple[Type[Any]]:
  """Returns callable's argument types."""
  argspec = inspect.getfullargspec(handler)

  def get_arg_type(arg: str):
    """Returns argument type or raises a user-friendly exception."""
    if arg not in argspec.annotations:
      raise ValueError(
          f'{arg} argument of a {handler} is missing a type annotation.'
      )
    return argspec.annotations[arg]

  return tuple(map(get_arg_type, argspec.args))


def match(*handlers: Union[Case[R], Callable[..., R]]) -> Callable[..., R]:
  """Creates a router applying appropriate function based on value's type.

  This is an implementation of https://en.wikipedia.org/wiki/Pattern_matching.

  Args:
    *handlers: Sequence of branches.

  Returns:
    A function that selects the earliest case whose arguments types are equal or
    base to the given value's type and applies its handler. If none of the cases
    match, a `TypeError` is raised.

  Examples:

    ```
    def visit_str(s: str):
      return reversed(s)

    matcher = match(
      visit_str,
      Case([int], lambda i: str(i)),
    )
    matcher(0)
    ```
  """

  cases = [
      handler
      if isinstance(handler, Case)
      else Case(_deduce_types(handler), handler)
      for handler in handlers
  ]

  def apply(*values: Any) -> R:
    for case in cases:
      if case.matches(*values):
        return case.handle(*values)

    value_types = tuple(type(value) for value in values)
    known_types = '\n'.join(str(case.kind) for case in cases)
    raise TypeError(
        f'{values} did not match any type pattern. Values have '
        'following types:\n'
        f'{value_types}\n'
        'Which does not match any of:\n'
        f'{known_types}'
    )

  return apply


def async_match(*handlers: Callable[..., R]) -> Callable[..., Awaitable[R]]:
  """`match` for async functions."""
  # Python type system doesn't differentiate regular and async functions enough.
  # `async def f(int) -> Cls:` is a `Callable[[int], Cls]` but actually returns
  # `Awaitable[Cls]`. This confuses pytype when it deduces `match` return type.

  return match(*handlers)



from lxm3.xm_cluster.config import Config
from lxm3.xm_cluster.executable_specs import CommandList
from lxm3.xm_cluster.executable_specs import Fileset
from lxm3.xm_cluster.executable_specs import ModuleName
from lxm3.xm_cluster.executable_specs import PythonPackage
from lxm3.xm_cluster.executable_specs import SingularityContainer
from lxm3.xm_cluster.executable_specs import UniversalPackage
from lxm3.xm_cluster.executables import Command
from lxm3.xm_cluster.executors import GridEngine
from lxm3.xm_cluster.executors import Local
from lxm3.xm_cluster.executors import SingularityOptions
from lxm3.xm_cluster.executors import Slurm
from lxm3.xm_cluster.experiment import ClusterExperiment
from lxm3.xm_cluster.experiment import create_experiment
from lxm3.xm_cluster.requirements import JobRequirements



# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A utility to predict IDs that would be assigned upon object creation.

This class helps to untie this chicken and egg problem. Sometimes to create an
object (for example WorkUnit) one may need to know its ID beforehand (for
example to generate a checkpoint path). But the ID would only be assigned by a
backend upon object creation. In ideal world we would rewrite the backend to
allow ID reservation. This module provides a temporary solution which does the
reservation on client-side. Following assumptions are made:
  * Ids are assigned sequentially by the backend, starting from some number.
  * Only one process at a time creates the objects. Any races are resolved only
    within that process.

Usage:
  predictor = Predictor()

  # Obtain the ID and asynchronously construct the object.
  next_id = predictor.reserve_id()
  job = Job(args={'checkpoint_path': f'/tmp/{next_id}'})

  # Wait untill all objects with smaller IDs are submitted.
  async with predictor.submit_id(next_id):
    # And submit it to the backend.
    submit(job)

If the submission fails, the sequence is considered broken and following calls
to the predictor would raise an error.
"""

import asyncio
import threading
from typing import AsyncIterable

import async_generator


class BrokenSequenceError(RuntimeError):
  """The ID would never be ready to submit."""


class Predictor:
  """Predicts IDs that would be assigned on object creation.

  This class is thread safe and async Python friendly. It must be constructed
  from inside asyncio event loop.
  """

  def __init__(self, next_id: int) -> None:
    """Initializes the predictor.

    Args:
      next_id: The first available ID that would be assigned to the next object.
    """
    self._next_id = next_id
    # We use threading.Lock to allow calling reserve_id from non async context.
    # Note that no long operations are done under this lock.
    self._next_id_lock = threading.Lock()

    self._is_broken = False
    self._last_created_id = next_id - 1
    self._last_created_id_condition = asyncio.Condition()

  def reserve_id(self) -> int:
    """Returns the next ID."""
    with self._next_id_lock:
      next_id = self._next_id
      self._next_id += 1
      return next_id

  @async_generator.asynccontextmanager
  async def submit_id(self, id_to_submit: int) -> AsyncIterable[None]:
    """Waits until the ID can be submitted and marks it as such.

    A context manager which would wait for all smaller IDs to be submitted on
    entry and marks it as submitted on exit. As a result all submissions are
    serialized in the correct order and receive the right ID from the backend.

    Args:
      id_to_submit: The id to be submitted.

    Yields:
      Yields when it is time to send the request to the backend.
    """
    async with self._last_created_id_condition:
      await self._last_created_id_condition.wait_for(
          lambda: self._is_broken or self._last_created_id == id_to_submit - 1
      )

      try:
        if self._is_broken:
          raise BrokenSequenceError(
              f'Id {id} would never be ready to submit as'
              ' submission of the previous one has failed'
          )
        yield
        self._last_created_id = id_to_submit
      except:
        self._is_broken = True
        raise
      finally:
        self._last_created_id_condition.notify_all()
